from functools import partial
from typing import (
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)
from lark import Tree, Token
from dataclasses import dataclass, field

from mamushi.formatting.comments import (
    add_leading_space_after_hashtag,
    settle_prefix,
)
from mamushi.formatting.nodes import (
    wrap_in_parentheses,
    is_atom_with_invisible_parens,
    ensure_visible,
    is_lpar_token,
    is_rpar_token,
)
from mamushi.formatting.strings import (
    normalize_string_quotes,
    is_multiline_string,
    fix_docstring,
    is_pragma,
)
from mamushi.parsing.pytree import Leaf, Node
from mamushi.formatting.lines import Line
from mamushi.parsing.tokens import SIMPLE_STATEMENTS, ASSIGNMENTS_SIGNS
from mamushi.parsing import tokens
import re

T = TypeVar("T")
TT = Union[Tree, Token]
LN = Union[Leaf, Node]

IMPORTS_TYPE = {tokens.IMPORT_FROM, tokens.IMPORT_NAME}

STATEMENT_TYPES = {
    tokens.FOR,
    tokens.IF,
    tokens.ELSE,
    tokens.ELIF,
    tokens.ASSERT_TOKEN,
}

ASSIGNMENTS = {
    tokens.DECLARATION,
    tokens.CONSTANT_DEF,
    tokens.ASSIGN,
    tokens.AUG_ASSIGN,
}


class Visitor(Generic[T]):
    def visit(self, node: LN) -> Iterator[T]:
        yield from getattr(self, f"visit_{node.type}", self.visit_default)(
            node
        )

    def visit_default(self, node: LN) -> Iterator[T]:
        if isinstance(node, Node):
            for child in node.children:
                yield from self.visit(child)


@dataclass
class LineGenerator(Visitor[Line]):
    """Generates reformatted Line objects.  Empty lines are not emitted.
    Note: destroys the tree it's visiting by mutating prefixes of its leaves
    in ways that will no longer stringify to valid Python code on the tree.
    """

    def __init__(self, line_length: int = 80) -> None:
        self.current_line: Line
        self.__post_init__()
        self.line_length = line_length

    def line(self, indent: int = 0) -> Iterator[Line]:
        """Generate a line.
        If the line is empty, only emit if it makes sense.
        If the line is too long, split it first and then generate.
        If any lines were generated, set up a new current_line.
        """
        if not self.current_line:
            self.current_line.depth += indent
            return  # Line is empty, don't emit. Creating a new one unnecessary.

        complete_line = self.current_line
        self.current_line = Line(depth=complete_line.depth + indent)
        yield complete_line

    def visit_default(self, node: LN) -> Iterator[Line]:
        if isinstance(node, Leaf):
            if node.type == tokens.STRING:
                node.value = normalize_string_quotes(node.value)
            if node.type not in tokens.WHITESPACE:
                self.current_line.append(node)
            if (
                node.type == tokens.STANDALONE_COMMENT
                and not self.current_line.bracket_tracker.any_open_brackets()
            ):
                yield from self.line()
        yield from super().visit_default(node)

    def visit_external_call(self, node: Node) -> Iterator[Line]:
        yield from self.visit_call(node)

    def visit_DOCSTRING(self, node: Leaf) -> Iterator[Line]:
        # ensure single/double quote consistency
        docstring = normalize_string_quotes(node.value)
        # don't further format header docstrings
        if node.parent and node.parent.type != tokens.MODULE:
            quote_char = docstring[0]
            # A natural way to remove the outer quotes is to do:
            #   docstring = docstring.strip(quote_char)
            # but that breaks on """""x""" (which is '""x').
            # So we actually need to remove the first character and the next two
            # characters but only if they are the same as the first.
            quote_len = 1 if docstring[1] != quote_char else 3
            docstring = docstring[quote_len:-quote_len]
            docstring_started_empty = not docstring
            indent = " " * 4 * self.current_line.depth

            if is_multiline_string(node):
                docstring = fix_docstring(docstring, indent)
            else:
                docstring = docstring.strip()

            if docstring:
                # Add some padding if the docstring starts / ends with a quote mark.
                if docstring[0] == quote_char:
                    docstring = " " + docstring
                if docstring[-1] == quote_char:
                    docstring += " "
                if docstring[-1] == "\\":
                    backslash_count = len(docstring) - len(
                        docstring.rstrip("\\")
                    )
                    if backslash_count % 2:
                        # Odd number of tailing backslashes, add some padding to
                        # avoid escaping the closing string quote.
                        docstring += " "
            elif not docstring_started_empty:
                docstring = " "

            # We could enforce triple quotes at this point.
            quote = quote_char * quote_len
            if quote_len == 3:
                # We need to find the length of the last line of the docstring
                # to find if we can add the closing quotes to the line without
                # exceeding the maximum line length.
                # If docstring is one line, then we need to add the length
                # of the indent, prefix, and starting quotes. Ending quotes are
                # handled later.
                lines = docstring.splitlines()
                last_line_length = len(lines[-1]) if docstring else 0

                if len(lines) == 1:
                    last_line_length += len(indent) + quote_len

                # If adding closing quotes would cause the last line to exceed
                # the maximum line length then put a line break before the
                # closing quotes
                if last_line_length + quote_len > self.line_length:
                    node.value = quote + docstring + "\n" + indent + quote
                else:
                    node.value = quote + docstring + quote
            else:
                node.value = quote + docstring + quote
        else:
            node.value = docstring
        yield from self.visit_default(node)

    def visit_COMMENT(self, node: Leaf) -> Iterator[Line]:
        any_open_brackets = (
            self.current_line.bracket_tracker.any_open_brackets()
        )
        node.value = add_leading_space_after_hashtag(node.value)
        settle_prefix(node)
        if is_pragma(node.value):
            node.type = tokens.PRAGMA
        self.current_line.append(node)
        if not any_open_brackets:
            # regular trailing comment
            yield from self.line()
        yield from super().visit_default(node)

    def visit_STANDALONE_COMMENT(self, node: Leaf) -> Iterator[Line]:
        node.value = re.sub(r"\n\n+", "\n\n", node.value)
        node.value = add_leading_space_after_hashtag(node.value)
        settle_prefix(node)
        if not self.current_line.bracket_tracker.any_open_brackets():
            yield from self.line()
        yield from self.visit_default(node)

    def visit__NEWLINE(self, node: Leaf) -> Iterator[Line]:
        nextl = next_leaf_add_newline(node)
        if nextl:
            nextl.prefix = "\n" + nextl.prefix
        else:
            yield from self.line()
        yield from super().visit_default(node)

    def visit__INDENT(self, node: Leaf) -> Iterator[Line]:
        """Increase indentation level, maybe yield a line."""
        yield from self.line(+1)
        yield from self.visit_default(node)

    def visit__DEDENT(self, node: Leaf) -> Iterator[Line]:
        """Decrease indentation level, maybe yield a line."""
        nextl = next_leaf_add_newline(node)
        if nextl:
            nextl.prefix = "\n" + nextl.prefix
        yield from self.line()

        # Finally, emit the dedent.
        yield from self.line(-1)

    def visit_body(self, node: Node) -> Iterator[Line]:
        yield from self.visit_default(node)

    def visit_call(self, node: Node) -> Iterator[Line]:
        """
        Handles calls depending on whether they're a standalone statement
        or part of a compound statement
        """
        if not node.parent or node.parent and node.parent.type == tokens.BODY:
            yield from self.line()
            yield from self.visit_default(node)
        else:
            for child in node.children:
                yield from self.visit(child)

    def visit_import(self, node: Node) -> Iterator[Line]:
        """Handles different import syntax"""
        normalize_invisible_parens(node, parens_after={"import"})
        for child in node.children:
            if child.type in IMPORTS_TYPE and child.prev_sibling is None:
                yield from self.line()
            yield from self.visit(child)

    def visit_stmt(
        self, node: Node, keywords: Set[str], parens: Set[str]
    ) -> Iterator[Line]:
        """Visit a statement."""
        normalize_invisible_parens(node, parens_after=parens)
        for child in node.children:
            if child.type in ({tokens.NAME} | tokens.DECLARATIONS | STATEMENT_TYPES) and child.value in keywords:  # type: ignore
                yield from self.line()

            yield from self.visit(child)

    def visit_simple_stmt(self, node: Node) -> Iterator[Line]:
        """A statement without nested statements."""
        if node.type in ASSIGNMENTS:
            normalize_invisible_parens(node, parens_after=ASSIGNMENTS_SIGNS)
        elif node.type in tokens.ASSERTS:
            normalize_invisible_parens(node, parens_after={"assert", ","})
        elif node.type == tokens.RETURN_STMT:
            normalize_invisible_parens(node, parens_after={"return"})
        elif node.type == tokens.EXPORT:
            normalize_invisible_parens(node, parens_after={"export", ":"})

        is_body_like = node.parent and (
            # TODO: handle this more cleanly
            node.parent.type not in tokens.BODIES
            # for single line body stmts
            or (
                node.parent.type == tokens.BODY
                and node.type in SIMPLE_STATEMENTS
                and not node.prev_sibling
            )
        )
        if is_body_like:
            yield from self.line(+1)
            yield from self.visit_default(node)
            yield from self.line(-1)

        else:
            yield from self.line()
            yield from self.visit_default(node)

    def visit_decorators(self, node: Node) -> Iterator[Line]:
        """Visit decorators."""
        for child in node.children:
            yield from self.line()
            yield from self.visit(child)

    def __post_init__(self) -> None:
        """You are in a twisty little maze of passages."""
        self.current_line = Line()
        v = self.visit_stmt
        self.visit_if_stmt = partial(
            v, keywords={"if", "else", "elif"}, parens={"if", "elif"}
        )
        self.visit_for_stmt = partial(v, keywords={"for", "else"}, parens={""})
        self.visit_function_sig = partial(v, keywords={"def"}, parens=set())

        for stmt in SIMPLE_STATEMENTS | ASSIGNMENTS:
            self.__setattr__(f"visit_{stmt}", self.visit_simple_stmt)


@dataclass
class EmptyLineTracker:
    """Provides a stateful method that returns the number of potential extra
    empty lines needed before and after the currently processed line.
    Note: this tracker works on lines that haven't been split yet.
    """

    previous_line: Optional[Line] = None
    previous_after: int = 0
    previous_defs: List[int] = field(default_factory=list)

    def maybe_empty_lines(self, current_line: Line) -> Tuple[int, int]:
        """Returns the number of extra empty lines before and after the `current_line`.
        This is for separating `def`, `async def` and `class` with extra empty lines
        (two on module-level), as well as providing an extra empty line after flow
        control keywords to make them more prominent.
        """
        before, after = self._maybe_empty_lines(current_line)
        before = (
            0 if self.previous_line is None else before - self.previous_after
        )
        self.previous_after = after
        self.previous_line = current_line
        return before, after

    def _maybe_empty_lines(self, current_line: Line) -> Tuple[int, int]:
        max_allowed = 2
        if current_line.leaves:
            # Consume the first leaf's extra newlines.
            first_leaf = current_line.leaves[0]
            before = first_leaf.prefix.count("\n")
            before = min(before, max_allowed)
            first_leaf.prefix = ""
        else:
            before = 0
        depth = current_line.depth
        while self.previous_defs and self.previous_defs[-1] >= depth:
            self.previous_defs.pop()
            before = (1 if depth else 2) - self.previous_after
        is_decorator = current_line.is_decorator
        if is_decorator or current_line.is_def:
            return self._maybe_empty_lines_for_class_or_def(
                current_line, before
            )

        if current_line.is_flow_control:
            return before, 1

        if current_line.is_pragma:
            return 0, 1

        if (
            self.previous_line
            and self.previous_line.is_import
            and not current_line.is_import
            and depth == self.previous_line.depth
        ):
            return (before or 1), 0

        return before, 0

    def _maybe_empty_lines_for_class_or_def(
        self, current_line: Line, before: int
    ) -> Tuple[int, int]:
        if not current_line.is_decorator:
            self.previous_defs.append(current_line.depth)
        if self.previous_line is None:
            # Don't insert empty lines before the first line in the file.
            return 0, 0

        if self.previous_line.is_decorator:
            return 0, 0

        if self.previous_line.depth < current_line.depth and (
            self.previous_line.is_def
        ):
            return 0, 0

        if (
            self.previous_line.is_comment
            and self.previous_line.depth == current_line.depth
            and before == 0
        ):
            return 0, 0

        else:
            newlines = 1 if current_line.depth else 2
        return newlines, 0


def next_leaf(node: Optional[LN]) -> Optional[Leaf]:
    """Return the first leaf that precedes `node`, if any."""
    while node:
        res = node.next_sibling
        if res:
            if isinstance(res, Leaf):
                return res

            try:
                return list(res.leaves())[0]

            except IndexError:
                return None

        node = node.parent
    return None


def normalize_invisible_parens(
    node: Node,
    parens_after: Set[str],
) -> None:
    """Make existing optional parentheses invisible or create new ones.

    `parens_after` is a set of string leaf values immediately after which parens
    should be put.

    Standardizes on visible parentheses for single-element tuples, and keeps
    existing visible parentheses for other tuples and generator expressions.
    """

    check_lpar = False
    for index, child in enumerate(list(node.children)):
        # Add parentheses around tuple assignments on lhs.
        if (
            index == 0
            and isinstance(child, Node)
            and child.type == tokens.MULTIPLE_ASSIGN
        ):
            check_lpar = True

        if check_lpar:
            if child.type == tokens.COND_EXEC:
                first_child = child.children[0]
                if (
                    isinstance(first_child, Node)
                    and isinstance(child, Node)
                    and maybe_make_parens_invisible_in_atom(
                        child.children[0],
                        parent=child,
                    )
                ):
                    wrap_in_parentheses(child, first_child, visible=False)

            elif node.type in {tokens.IMPORT, tokens.EXPORT}:
                _normalize_import_from(node, child, index)
                break

            elif not (isinstance(child, Leaf)):
                wrap_in_parentheses(node, child, visible=False)

        check_lpar = isinstance(child, Leaf) and (child.value in parens_after)


def _normalize_import_from(parent: Node, child: LN, index: int) -> None:
    # "import from" nodes store parentheses directly as part of
    # the statement
    if (
        is_lpar_token(child)
        and isinstance(child, Leaf)
        and isinstance(parent.children[-1], Leaf)
        and is_rpar_token(parent.children[-1])
    ):
        # make parentheses invisible
        child.value = ""
        parent.children[-1].value = ""
    elif child.type != tokens.STAR:
        # insert invisible parentheses
        parent.insert_child(index, Leaf(tokens.LPAR, ""))
        parent.append_child(Leaf(tokens.RPAR, ""))


def maybe_make_parens_invisible_in_atom(
    node: LN,
    parent: LN,
    remove_brackets_around_comma: bool = False,
) -> bool:
    """If it's safe, make the parens in the atom `node` invisible, recursively.
    Additionally, remove repeated, adjacent invisible parens from the atom `node`
    as they are redundant.

    Returns whether the node should itself be wrapped in invisible parentheses.
    """

    first = node.children[0]
    last = node.children[-1]
    if (
        isinstance(first, Leaf)
        and first.type == tokens.LPAR
        and first.value == "("
        and isinstance(last, Leaf)
        and last.type == tokens.RPAR
        and last.value == ")"
    ):
        middle = node.children[1]
        # make parentheses invisible
        first.value = ""
        last.value = ""
        maybe_make_parens_invisible_in_atom(
            middle,
            parent=parent,
            remove_brackets_around_comma=remove_brackets_around_comma,
        )

        if is_atom_with_invisible_parens(middle):
            # Strip the invisible parens from `middle` by replacing
            # it with the child in-between the invisible parens
            middle.replace(middle.children[1])

        return False

    return True


def get_next_non_whitespace_leaf(node: Leaf) -> Optional[Leaf]:
    nextl = next_leaf(node)
    while nextl and nextl.type in tokens.WHITESPACE:
        nextl = next_leaf(nextl)
    return nextl


def next_leaf_add_newline(node: Leaf) -> Optional[Leaf]:
    if node.value.strip(" \t").endswith("\n\n"):
        # If user inserted multiple blank lines, we reduce to 1
        return get_next_non_whitespace_leaf(node)
    return None



"""Builds on top of nodes.py to track brackets."""

from dataclasses import dataclass, field
from typing import Dict, Iterable, List, Optional, Tuple, Union
from mamushi.parsing import tokens
from typing import Final
from mamushi.parsing.pytree import Leaf, Node


# types

LN = Union[Leaf, Node]
Depth = int
LeafID = int
NodeType = str
Priority = int


COMPREHENSION_PRIORITY: Final = 20
COMMA_PRIORITY: Final = 18
TERNARY_PRIORITY: Final = 16
LOGIC_PRIORITY: Final = 14
STRING_PRIORITY: Final = 12
COMPARATOR_PRIORITY: Final = 10
MATH_PRIORITIES: Final = {
    tokens.VBAR: 9,
    tokens.CARET: 8,
    tokens.AMPERSAND: 7,
    tokens.LEFTSHIFT: 6,
    tokens.RIGHTSHIFT: 6,
    tokens.PLUS: 5,
    tokens.MINUS: 5,
    tokens.STAR: 4,
    tokens.SLASH: 4,
    tokens.PERCENT: 4,
    tokens.AT: 4,
    tokens.TILDE: 3,
    tokens.DOUBLESTAR: 2,
}
DOT_PRIORITY: Final = 1


class BracketMatchError(Exception):
    """Raised when an opening bracket is unable to be matched to a closing bracket."""


@dataclass
class BracketTracker:
    """Keeps track of brackets on a line."""

    depth: int = 0
    bracket_match: Dict[Tuple[Depth, NodeType], Leaf] = field(
        default_factory=dict
    )
    delimiters: Dict[LeafID, Priority] = field(default_factory=dict)
    previous: Optional[Leaf] = None
    _for_loop_depths: List[int] = field(default_factory=list)
    _lambda_argument_depths: List[int] = field(default_factory=list)
    invisible: List[Leaf] = field(default_factory=list)

    def mark(self, leaf: Leaf) -> None:
        """Mark `leaf` with bracket-related metadata. Keep track of delimiters.

        All leaves receive an int `bracket_depth` field that stores how deep
        within brackets a given leaf is. 0 means there are no enclosing brackets
        that started on this line.

        If a leaf is itself a closing bracket, it receives an `opening_bracket`
        field that it forms a pair with. This is a one-directional link to
        avoid reference cycles.

        If a leaf is a delimiter (a token on which Black can split the line if
        needed) and it's on depth 0, its `id()` is stored in the tracker's
        `delimiters` field.
        """
        if leaf.type == tokens.COMMENT:
            return

        self.maybe_decrement_after_for_loop_variable(leaf)
        if leaf.type in tokens.CLOSING_BRACKETS:
            self.depth -= 1
            try:
                opening_bracket = self.bracket_match.pop(
                    (self.depth, leaf.type)
                )
            except KeyError as e:
                raise BracketMatchError(
                    "Unable to match a closing bracket to the following opening"
                    f" bracket: {leaf}"
                ) from e
            leaf.opening_bracket = opening_bracket
            if not leaf.value:
                self.invisible.append(leaf)
        leaf.bracket_depth = self.depth
        if self.depth == 0:
            delim = is_split_before_delimiter(leaf, self.previous)
            if delim and self.previous is not None:
                self.delimiters[id(self.previous)] = delim
            else:
                delim = is_split_after_delimiter(leaf, self.previous)
                if delim:
                    self.delimiters[id(leaf)] = delim
        if leaf.type in tokens.OPENING_BRACKETS:
            self.bracket_match[self.depth, tokens.BRACKET_MAP[leaf.type]] = (
                leaf
            )
            self.depth += 1
            if not leaf.value:
                self.invisible.append(leaf)
        self.previous = leaf
        self.maybe_increment_for_loop_variable(leaf)

    def any_open_brackets(self) -> bool:
        """Return True if there is an yet unmatched open bracket on the line."""
        return bool(self.bracket_match)

    def max_delimiter_priority(
        self, exclude: Iterable[LeafID] = ()
    ) -> Priority:
        """Return the highest priority of a delimiter found on the line.

        Values are consistent with what `is_split_*_delimiter()` return.
        Raises ValueError on no delimiters.
        """
        return max(v for k, v in self.delimiters.items() if k not in exclude)

    def delimiter_count_with_priority(self, priority: Priority = 0) -> int:
        """Return the number of delimiters with the given `priority`.

        If no `priority` is passed, defaults to max priority on the line.
        """
        if not self.delimiters:
            return 0

        priority = priority or self.max_delimiter_priority()
        return sum(1 for p in self.delimiters.values() if p == priority)

    def maybe_increment_for_loop_variable(self, leaf: Leaf) -> bool:
        """In a for loop, or comprehension, the variables are often unpacks.

        To avoid splitting on the comma in this situation, increase the depth of
        tokens between `for` and `in`.
        """
        if leaf.type == tokens.FOR and leaf.value == "for":
            self.depth += 1
            self._for_loop_depths.append(self.depth)
            return True

        return False

    def maybe_decrement_after_for_loop_variable(self, leaf: Leaf) -> bool:
        """See `maybe_increment_for_loop_variable` above for explanation."""
        if (
            self._for_loop_depths
            and self._for_loop_depths[-1] == self.depth
            and leaf.type == tokens.IN
            and leaf.value == "in"
        ):
            self.depth -= 1
            self._for_loop_depths.pop()
            return True

        return False


def is_split_after_delimiter(
    leaf: Leaf, previous: Optional[Leaf] = None
) -> Priority:
    """Return the priority of the `leaf` delimiter, given a line break after it.

    The delimiter priorities returned here are from those delimiters that would
    cause a line break after themselves.

    Higher numbers are higher priority.
    """
    if leaf.type == tokens.COMMA:
        return COMMA_PRIORITY

    return 0


def is_split_before_delimiter(
    leaf: Leaf, previous: Optional[Leaf] = None
) -> Priority:
    """Return the priority of the `leaf` delimiter, given a line break before it.

    The delimiter priorities returned here are from those delimiters that would
    cause a line break before themselves.

    Higher numbers are higher priority.
    """

    if (
        leaf.type == tokens.DOT
        and leaf.parent
        and leaf.parent.type != tokens.IMPORT
        and (previous is None or previous.type in tokens.CLOSING_BRACKETS)
    ):
        return DOT_PRIORITY

    if leaf.type in tokens.MATH_OPERATORS and leaf.parent:
        return MATH_PRIORITIES[leaf.type]

    if leaf.type in tokens.COMPARATORS:
        return COMPARATOR_PRIORITY

    if (
        leaf.type == tokens.STRING
        and previous is not None
        and previous.type == tokens.STRING
    ):
        return STRING_PRIORITY

    if (
        leaf.value == "for"
        and leaf.parent
        and leaf.parent.type == tokens.FOR_STMT
    ):
        if not isinstance(leaf.prev_sibling, Leaf):
            return COMPREHENSION_PRIORITY

    if (
        leaf.value == "if"
        and leaf.parent
        and leaf.parent.type in {tokens.IF_STMT, tokens.COND_EXEC}
    ):
        return COMPREHENSION_PRIORITY

    if (
        leaf.value in {"if", "else"}
        and leaf.parent
        and leaf.parent.type == tokens.TERNARY
    ):
        return TERNARY_PRIORITY

    if leaf.value in tokens.LOGIC_OPERATORS and leaf.parent:
        return LOGIC_PRIORITY

    return 0



from typing import Iterator
from pathlib import Path

VYPER_EXTENSIONS = {".vy", ".vyi"}
BLACKLISTED_DIRECTORIES = {
    "build",
    "buck-out",
    ".nox",
    "venv",
    ".direnv",
    ".eggs",
    "__pypackages__",
    "build",
    ".ipynb_checkpoints",
    "dist",
    "_build",
    ".git",
    ".hg",
    ".mypy_cache",
    ".tox",
    ".venv",
    ".idea",
}


def gen_vyper_files_in_dir(path: Path) -> Iterator[Path]:
    for child in path.iterdir():
        if child.is_dir():
            if child.name in BLACKLISTED_DIRECTORIES:
                continue

            yield from gen_vyper_files_in_dir(child)

        elif child.suffix in VYPER_EXTENSIONS:
            yield child



import pytest
from mamushi import gen_vyper_files_in_dir


def test_gen_python_files_in_dir(tmp_path):
    valid_files = []
    for i, dir in enumerate(["build"] + list("abc")):
        fake_dir = tmp_path / dir
        fake_dir.mkdir()
        for j, ext in enumerate([".py", ".vy"]):
            fake_file = fake_dir / ("test" + ext)
            fake_file.write_text("test")
            if i * j:
                valid_files.append(fake_file)
    found_files = gen_vyper_files_in_dir(tmp_path)
    assert set(valid_files) == set(found_files)



# Copyright 2006 Google, Inc. All Rights Reserved.
# Licensed to PSF under a Contributor Agreement.

"""
Python parse tree definitions.

This is a very concrete parse tree; we need to keep every token and
even the comments and whitespace between tokens.

There's also a pattern matching implementation here.
"""

# mypy: allow-untyped-defs, allow-incomplete-defs

from typing import (
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Text,
    Tuple,
    TypeVar,
    Union,
    Set,
    Iterable,
)
from lark import Tree, Token

__author__ = "Guido van Rossum <guido@python.org>"

import sys
from io import StringIO

HUGE: int = 0x7FFFFFFF  # maximum repeat count, default max

_type_reprs: Dict[int, Union[Text, int]] = {}


_P = TypeVar("_P", bound="Base")

NL = Union["Node", "Leaf"]
Context = Tuple[Text, Tuple[int, int]]
RawNode = Tuple[int, Optional[Text], Optional[Context], Optional[List[NL]]]


class Base(object):
    """
    Abstract base class for Node and Leaf.

    This provides some default functionality and boilerplate using the
    template pattern.

    A node may be a subnode of at most one parent.
    """

    # Default values for instance variables
    type: str
    parent: Optional["Node"] = None  # Parent node pointer, or None
    children: List[NL]  # List of subnodes
    was_changed: bool = False
    was_checked: bool = False

    def __new__(cls, *args, **kwds):
        """Constructor that prevents Base from being instantiated."""
        assert cls is not Base, "Cannot instantiate Base"
        return object.__new__(cls)

    def __eq__(self, other: Any) -> bool:
        """
        Compare two nodes for equality.

        This calls the method _eq().
        """
        if self.__class__ is not other.__class__:
            return NotImplemented
        return self._eq(other)

    @property
    def prefix(self) -> Text:
        raise NotImplementedError

    def _eq(self: _P, other: _P) -> bool:
        """
        Compare two nodes for equality.

        This is called by __eq__ and __ne__.  It is only called if the two nodes
        have the same type.  This must be implemented by the concrete subclass.
        Nodes should be considered equal if they have the same structure,
        ignoring the prefix string and other context information.
        """
        raise NotImplementedError

    def __deepcopy__(self: _P, memo: Any) -> _P:
        return self.clone()

    def clone(self: _P) -> _P:
        """
        Return a cloned (deep) copy of self.

        This must be implemented by the concrete subclass.
        """
        raise NotImplementedError

    def post_order(self) -> Iterator[NL]:
        """
        Return a post-order iterator for the tree.

        This must be implemented by the concrete subclass.
        """
        raise NotImplementedError

    def pre_order(self) -> Iterator[NL]:
        """
        Return a pre-order iterator for the tree.

        This must be implemented by the concrete subclass.
        """
        raise NotImplementedError

    def replace(self, new: Union[NL, List[NL]]) -> None:
        """Replace this node with a new one in the parent."""
        assert self.parent is not None, str(self)
        assert new is not None
        if not isinstance(new, list):
            new = [new]
        l_children = []
        found = False
        for ch in self.parent.children:
            if ch is self:
                assert not found, (self.parent.children, self, new)
                if new is not None:
                    l_children.extend(new)
                found = True
            else:
                l_children.append(ch)
        assert found, (self.children, self, new)
        self.parent.children = l_children
        self.parent.changed()
        self.parent.invalidate_sibling_maps()
        for x in new:
            x.parent = self.parent
        self.parent = None

    def get_lineno(self) -> Optional[int]:
        """Return the line number which generated the invocant node."""
        node = self
        while not isinstance(node, Leaf):
            if not node.children:
                return None
            node = node.children[0]
        return node.lineno

    def changed(self) -> None:
        if self.was_changed:
            return
        if self.parent:
            self.parent.changed()
        self.was_changed = True

    def remove(self) -> Optional[int]:
        """
        Remove the node from the tree. Returns the position of the node in its
        parent's children before it was removed.
        """
        if self.parent:
            for i, node in enumerate(self.parent.children):
                if node is self:
                    del self.parent.children[i]
                    self.parent.changed()
                    self.parent.invalidate_sibling_maps()
                    self.parent = None
                    return i
        return None

    @property
    def next_sibling(self) -> Optional[NL]:
        """
        The node immediately following the invocant in their parent's children
        list. If the invocant does not have a next sibling, it is None
        """
        if self.parent is None:
            return None

        if self.parent.next_sibling_map is None:
            self.parent.update_sibling_maps()
        assert self.parent.next_sibling_map is not None
        return self.parent.next_sibling_map[id(self)]

    @property
    def prev_sibling(self) -> Optional[NL]:
        """
        The node immediately preceding the invocant in their parent's children
        list. If the invocant does not have a previous sibling, it is None.
        """
        if self.parent is None:
            return None

        if self.parent.prev_sibling_map is None:
            self.parent.update_sibling_maps()
        assert self.parent.prev_sibling_map is not None
        return self.parent.prev_sibling_map[id(self)]

    def leaves(self) -> Iterator["Leaf"]:
        for child in self.children:
            yield from child.leaves()

    def depth(self) -> int:
        if self.parent is None:
            return 0
        return 1 + self.parent.depth()

    def get_suffix(self) -> Text:
        """
        Return the string immediately following the invocant node. This is
        effectively equivalent to node.next_sibling.prefix
        """
        next_sib = self.next_sibling
        if next_sib is None:
            return ""
        prefix = next_sib.prefix
        return prefix


class Node(Base):
    """Concrete implementation for interior nodes."""

    fixers_applied: Optional[List[Any]]
    used_names: Optional[Set[Text]]

    def __init__(
        self,
        type: str,
        children: List[NL],
        context: Optional[Any] = None,
        prefix: Optional[Text] = None,
        fixers_applied: Optional[List[Any]] = None,
    ) -> None:
        """
        Initializer.

        Takes a type constant, a sequence of
        child nodes, and an optional context keyword argument.

        As a side effect, the parent pointers of the children are updated.
        """
        self.type = type
        self.children = list(children)
        for ch in self.children:
            assert ch.parent is None, repr(ch)
            ch.parent = self
        self.invalidate_sibling_maps()
        if prefix is not None:
            self.prefix = prefix
        if fixers_applied:
            self.fixers_applied = fixers_applied[:]
        else:
            self.fixers_applied = None

    def __repr__(self) -> Text:
        """Return a canonical string representation."""
        assert self.type is not None
        return "%s(%s, %r)" % (
            self.__class__.__name__,
            self.type,
            self.children,
        )

    def __str__(self) -> Text:
        """
        Return a pretty string representation.

        This reproduces the input source exactly.
        """
        return "".join(map(str, self.children))

    def _eq(self, other: Base) -> bool:
        """Compare two nodes for equality."""
        return (self.type, self.children) == (other.type, other.children)

    def clone(self) -> "Node":
        assert self.type is not None
        """Return a cloned (deep) copy of self."""
        return Node(
            self.type,
            [ch.clone() for ch in self.children],
            fixers_applied=self.fixers_applied,
        )

    def post_order(self) -> Iterator[NL]:
        """Return a post-order iterator for the tree."""
        for child in self.children:
            yield from child.post_order()
        yield self

    def pre_order(self) -> Iterator[NL]:
        """Return a pre-order iterator for the tree."""
        yield self
        for child in self.children:
            yield from child.pre_order()

    @property
    def prefix(self) -> Text:
        """
        The whitespace and comments preceding this node in the input.
        """
        if not self.children:
            return ""
        return self.children[0].prefix

    @prefix.setter
    def prefix(self, prefix: Text) -> None:
        if self.children:
            self.children[0].prefix = prefix

    def set_child(self, i: int, child: NL) -> None:
        """
        Equivalent to 'node.children[i] = child'. This method also sets the
        child's parent attribute appropriately.
        """
        child.parent = self
        self.children[i].parent = None
        self.children[i] = child
        self.changed()
        self.invalidate_sibling_maps()

    def insert_child(self, i: int, child: NL) -> None:
        """
        Equivalent to 'node.children.insert(i, child)'. This method also sets
        the child's parent attribute appropriately.
        """
        child.parent = self
        self.children.insert(i, child)
        self.changed()
        self.invalidate_sibling_maps()

    def append_child(self, child: NL) -> None:
        """
        Equivalent to 'node.children.append(child)'. This method also sets the
        child's parent attribute appropriately.
        """
        child.parent = self
        self.children.append(child)
        self.changed()
        self.invalidate_sibling_maps()

    def invalidate_sibling_maps(self) -> None:
        self.prev_sibling_map: Optional[Dict[int, Optional[NL]]] = None
        self.next_sibling_map: Optional[Dict[int, Optional[NL]]] = None

    def update_sibling_maps(self) -> None:
        _prev: Dict[int, Optional[NL]] = {}
        _next: Dict[int, Optional[NL]] = {}
        self.prev_sibling_map = _prev
        self.next_sibling_map = _next
        previous: Optional[NL] = None
        for current in self.children:
            _prev[id(current)] = previous
            _next[id(previous)] = current
            previous = current
        _next[id(current)] = None


class Leaf(Base):
    """Concrete implementation for leaf nodes."""

    # Default values for instance variables
    value: Text
    fixers_applied: List[Any]
    bracket_depth: int
    # Changed later in brackets.py
    opening_bracket: Optional["Leaf"] = None
    used_names: Optional[Set[Text]]
    _prefix = ""  # Whitespace and comments preceding this token in the input
    lineno: int = 0  # Line where this token starts in the input
    column: int = 0  # Column where this token starts in the input

    def __init__(
        self,
        type: str,
        value: Text,
        context: Optional[Context] = None,
        prefix: Optional[Text] = None,
        fixers_applied: List[Any] = [],
        opening_bracket: Optional["Leaf"] = None,
    ) -> None:
        """
        Initializer.

        Takes a type constant, a string value, and an
        optional context keyword argument.
        """

        if context is not None:
            self._prefix, (self.lineno, self.column) = context
        self.type = type
        self.value = value
        if prefix is not None:
            self._prefix = prefix
        self.fixers_applied: Optional[List[Any]] = fixers_applied[:]
        self.children = []
        self.opening_bracket = opening_bracket

    def __repr__(self) -> str:
        """Return a canonical string representation."""

        assert self.type is not None
        return "%s(%s, %r)" % (
            self.__class__.__name__,
            self.type,
            self.value,
        )

    def __str__(self) -> Text:
        """
        Return a pretty string representation.

        This reproduces the input source exactly.
        """
        return self._prefix + str(self.value)

    def _eq(self, other: "Leaf") -> bool:
        """Compare two nodes for equality."""
        return (self.type, self.value) == (other.type, other.value)

    def clone(self) -> "Leaf":
        assert self.type is not None
        """Return a cloned (deep) copy of self."""
        return Leaf(
            self.type,
            self.value,
            (self.prefix, (self.lineno, self.column)),
            fixers_applied=self.fixers_applied,
        )

    def leaves(self) -> Iterator["Leaf"]:
        yield self

    def post_order(self) -> Iterator["Leaf"]:
        """Return a post-order iterator for the tree."""
        yield self

    def pre_order(self) -> Iterator["Leaf"]:
        """Return a pre-order iterator for the tree."""
        yield self

    @property
    def prefix(self) -> Text:
        """
        The whitespace and comments preceding this token in the input.
        """
        return self._prefix

    @prefix.setter
    def prefix(self, prefix: Text) -> None:
        self.changed()
        self._prefix = prefix



"Provides Indentation services for languages with indentation similar to Python"

from abc import ABC, abstractmethod
from typing import List, Iterator, Tuple
from mamushi.parsing.tokens import (
    OPENING_BRACKETS,
    CLOSING_BRACKETS,
    NEWLINE,
    INDENT,
    DEDENT,
)

from lark.exceptions import LarkError
from lark.lark import PostLex
from lark.lexer import Token


class DedentError(LarkError):
    pass


class Indenter(PostLex, ABC):
    paren_level: int
    indent_level: List[int]
    processed_newlines: List[Token]

    def __init__(self) -> None:
        self.paren_level = 0
        self.indent_level = [0]
        self.processed_newlines = []
        assert self.tab_len > 0

    @staticmethod
    def create_dent_on_next_line(
        type: str, value: str, newline: Token, lines: int
    ):
        endline = newline.end_line if newline.end_line else 0
        return Token(
            type,
            value,
            newline.start_pos,
            endline + lines,
            0,
            endline + lines,
            0,
            0,
        )

    def split_into_dedents(self, line: str) -> List[str]:
        res = []
        prev_tab_len = self.indent_level[-1]
        for i, element in enumerate(line.split("\n")[1:]):
            pre_comment = element.split("#")[0]
            cur_tab_len = (
                pre_comment.count(" ") + pre_comment.count("\t") * self.tab_len
            )
            if (cur_tab_len < prev_tab_len) or i == 0:
                res.append("\n" + element)
                if len(element.strip()) > 0:
                    prev_tab_len = cur_tab_len
                else:
                    prev_tab_len = prev_tab_len - self.tab_len
                    # the newline will get added to a newline token and is gone
                    # need to add it back as prefix to next comment
                    res[-1] += "\n"
            else:
                res[-1] += "\n" + element
        # if the final element is whitespace, merge it with penultimate
        if not res[-1].strip() and len(res) > 1:
            res[-2] += res[-1]
            res.pop()

        return res

    @staticmethod
    def split_into_indents(line: str) -> Tuple[str, str]:
        if "#" not in line:
            return "", line
        nl, ind = line.split("#", 1)
        return nl, "#" + ind

    def _get_indent(self, token):
        indent_str = token.rsplit("\n", 1)[1]  # Tabs and spaces
        return indent_str.count(" ") + indent_str.count("\t") * self.tab_len

    def handle_NL(self, token: Token) -> Iterator[Token]:
        if self.paren_level > 0:
            return

        self.processed_newlines.append(token)  # Tabs and spaces
        indent = self._get_indent(token)

        if indent > self.indent_level[-1]:
            # nl, ind = self.split_into_indents(token.value)
            yield Token.new_borrow_pos(self.NL_type, "", token)
            yield self.create_dent_on_next_line(
                self.INDENT_type, token.value, token, 1
            )
            self.indent_level.append(indent)
        elif indent < self.indent_level[-1]:
            dedent_captions = self.split_into_dedents(token)
            index = 0
            newline = False  # whether we've created the leading newline or not
            newline_caption = ""
            while indent < self.indent_level[-1]:
                caption = (
                    dedent_captions[index]
                    if index < len(dedent_captions)
                    else ""
                )
                index += 1

                if not newline:
                    # as long as we haven't hit the dedent, we add the comments
                    # to the newline
                    cmt_indent = (
                        self._get_indent(caption.split("#")[0])
                        if "#" in caption
                        else indent
                    )
                    if cmt_indent >= self.indent_level[-1]:
                        newline_caption += caption
                        continue

                    yield Token.new_borrow_pos(
                        self.NL_type, newline_caption.rstrip(), token
                    )
                    newline = True
                self.indent_level.pop()
                yield self.create_dent_on_next_line(
                    self.DEDENT_type, caption, token, 1
                )

            if indent != self.indent_level[-1]:
                raise DedentError(
                    "Unexpected dedent to column %s. Expected dedent to %s"
                    % (indent, self.indent_level[-1])
                )
        else:
            yield token

    def _process(self, stream):
        for token in stream:
            if token.type == self.NL_type:
                yield from self.handle_NL(token)
            else:
                yield token

            if token.type in self.OPEN_PAREN_types:
                self.paren_level += 1
            elif token.type in self.CLOSE_PAREN_types:
                self.paren_level -= 1
                assert self.paren_level >= 0

        while len(self.indent_level) > 1:
            self.indent_level.pop()
            yield Token(self.DEDENT_type, "")

        assert self.indent_level == [0], self.indent_level

    def process(self, stream):
        self.paren_level = 0
        self.indent_level = [0]
        self.processed_newlines = []
        return self._process(stream)

    # XXX Hack for ContextualLexer. Maybe there's a more elegant solution?
    @property
    def always_accept(self):
        return (self.NL_type,)

    @property
    @abstractmethod
    def NL_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def OPEN_PAREN_types(self) -> List[str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def CLOSE_PAREN_types(self) -> List[str]:
        raise NotImplementedError()

    @property
    @abstractmethod
    def INDENT_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def DEDENT_type(self) -> str:
        raise NotImplementedError()

    @property
    @abstractmethod
    def tab_len(self) -> int:
        raise NotImplementedError()


class PythonIndenter(Indenter):
    NL_type = NEWLINE
    OPEN_PAREN_types = list(OPENING_BRACKETS)
    CLOSE_PAREN_types = list(CLOSING_BRACKETS)
    INDENT_type = INDENT
    DEDENT_type = DEDENT
    tab_len = 4



"""
Simple formatting on strings. Further string formatting code is in trans.py.
"""

import re
import sys
from functools import lru_cache
from typing import List, Pattern, Final

from mamushi.parsing.pytree import Leaf

FIRST_NON_WHITESPACE_RE: Final = re.compile(r"\s*\t+\s*(\S)")


def has_triple_quotes(string: str) -> bool:
    """
    Returns:
        True iff @string starts with three quotation characters.
    """
    return string[:3] in {'"""', "'''"}


def is_multiline_string(leaf: Leaf) -> bool:
    """Return True if `leaf` is a multiline string that actually spans many lines."""
    return has_triple_quotes(leaf.value) and "\n" in leaf.value


def sub_twice(regex: Pattern[str], replacement: str, original: str) -> str:
    """Replace `regex` with `replacement` twice on `original`.

    This is used by string normalization to perform replaces on
    overlapping matches.
    """
    return regex.sub(replacement, regex.sub(replacement, original))


def lines_with_leading_tabs_expanded(s: str) -> List[str]:
    """
    Splits string into lines and expands only leading tabs (following the normal
    Python rules)
    """
    lines = []
    for line in s.splitlines():
        # Find the index of the first non-whitespace character after a string of
        # whitespace that includes at least one tab
        match = FIRST_NON_WHITESPACE_RE.match(line)
        if match:
            first_non_whitespace_idx = match.start(1)

            lines.append(
                line[:first_non_whitespace_idx].expandtabs()
                + line[first_non_whitespace_idx:]
            )
        else:
            lines.append(line)
    return lines


def fix_docstring(docstring: str, prefix: str) -> str:
    # https://www.python.org/dev/peps/pep-0257/#handling-docstring-indentation
    if not docstring:
        return ""
    lines = lines_with_leading_tabs_expanded(docstring)
    # Determine minimum indentation (first line doesn't count):
    indent = sys.maxsize
    for line in lines[1:]:
        stripped = line.lstrip()
        if stripped:
            indent = min(indent, len(line) - len(stripped))
    # Remove indentation (first line is special):
    trimmed = [lines[0].strip()]
    if indent < sys.maxsize:
        last_line_idx = len(lines) - 2
        for i, line in enumerate(lines[1:]):
            stripped_line = line[indent:].rstrip()
            if stripped_line or i == last_line_idx:
                trimmed.append(prefix + stripped_line)
            else:
                trimmed.append("")
    return "\n".join(trimmed)


# Re(gex) does actually cache patterns internally but this still improves
# performance on a long list literal of strings by 5-9% since lru_cache's
# caching overhead is much lower.
@lru_cache(maxsize=64)
def _cached_compile(pattern: str) -> Pattern[str]:
    return re.compile(pattern)


def normalize_string_quotes(s: str) -> str:
    """Prefer double quotes but only if it doesn't cause more escaping.

    Adds or removes backslashes as appropriate. Doesn't parse and fix
    strings nested in f-strings.
    """

    if s[:3] == '"""':
        return s

    elif s[:3] == "'''":
        orig_quote = "'''"
        new_quote = '"""'
    elif s[0] == '"':
        orig_quote = '"'
        new_quote = "'"
    else:
        orig_quote = "'"
        new_quote = '"'
    first_quote_pos = s.find(orig_quote)
    if first_quote_pos == -1:
        return s  # There's an internal error

    prefix = s[:first_quote_pos]
    unescaped_new_quote = _cached_compile(rf"(([^\\]|^)(\\\\)*){new_quote}")
    escaped_new_quote = _cached_compile(rf"([^\\]|^)\\((?:\\\\)*){new_quote}")
    escaped_orig_quote = _cached_compile(
        rf"([^\\]|^)\\((?:\\\\)*){orig_quote}"
    )
    body = s[first_quote_pos + len(orig_quote) : -len(orig_quote)]

    # remove unnecessary escapes
    new_body = sub_twice(escaped_new_quote, rf"\1\2{new_quote}", body)
    if body != new_body:
        # Consider the string without unnecessary escapes as the original
        body = new_body
        s = f"{prefix}{orig_quote}{body}{orig_quote}"
    new_body = sub_twice(escaped_orig_quote, rf"\1\2{orig_quote}", new_body)
    new_body = sub_twice(unescaped_new_quote, rf"\1\\{new_quote}", new_body)

    if new_quote == '"""' and new_body[-1:] == '"':
        # edge case:
        new_body = new_body[:-1] + '\\"'
    orig_escape_count = body.count("\\")
    new_escape_count = new_body.count("\\")
    if new_escape_count > orig_escape_count:
        return s  # Do not introduce more escaping

    if new_escape_count == orig_escape_count and orig_quote == '"':
        return s  # Prefer double quotes

    return f"{prefix}{new_quote}{new_body}{new_quote}"


def is_pragma(string: str) -> bool:
    version = re.match(r"#\s*@version\s*(\d\.?)*", string)
    return (version is not None) and (version.group(0) == string.strip())


def remove_double_spaces(string: str) -> str:
    return re.sub(r"\s+", " ", string)



from mamushi.formatting.linegen import LineGenerator, EmptyLineTracker
from mamushi.formatting.lines import Line, split_line
from mamushi.parsing.pytree import Node


def format_tree(ast: Node, max_line_length: int = 80) -> str:
    lg = LineGenerator(max_line_length)
    elt = EmptyLineTracker()
    empty_line = Line()
    after = 0
    dst_contents = []

    for current_line in lg.visit(ast):
        dst_contents.append(str(empty_line) * after)
        before, after = elt.maybe_empty_lines(current_line)
        dst_contents.append(str(empty_line) * before)
        if "# nosplit" in str(current_line):
            dst_contents.append(str(current_line))
        else:
            for line in split_line(current_line, line_length=max_line_length):
                dst_contents.append(str(line))

    return "".join(dst_contents)



from typing import Final
from typing import Optional, TypeVar, Union
from mamushi.parsing import tokens
from mamushi.parsing.pytree import Node, Leaf

T = TypeVar("T")
LN = Union[Leaf, Node]
ALWAYS_NO_SPACE: Final = tokens.CLOSING_BRACKETS | {
    tokens.COMMA,
    tokens.COLON,
    tokens.STANDALONE_COMMENT,
    tokens.DOCSTRING,
}


def whitespace(leaf: Leaf) -> str:
    """
    Handle adding whitespace or not before tokens
    Adapted from black
    TODO: complexity too high, needs to be refactored
    """
    NO: Final = ""
    SPACE: Final = " "
    DOUBLESPACE: Final = "  "
    t = leaf.type
    p = leaf.parent

    if t in ALWAYS_NO_SPACE:
        return NO

    if t == tokens.COMMENT:
        return DOUBLESPACE

    if not p:
        return SPACE

    if p.type == tokens.DECORATOR:
        return NO

    prev = leaf.prev_sibling

    if not prev:
        prevp = preceding_leaf(p)
        if not prevp or prevp.type in tokens.OPENING_BRACKETS:
            return NO
        if prevp.type == tokens.TILDE:
            return NO
        # kwargs with expr assignment
        if (
            prevp.type == tokens.EQUAL
            and prevp.parent
            and prevp.parent.type == tokens.KWARG
        ):
            return NO

        elif (
            prevp.parent
            and prevp.parent.type in tokens.UNARY
            and prevp.type in tokens.MATH_OPERATORS
        ):
            return NO

        elif (
            prevp.parent
            and prevp.parent.parent
            and prevp.parent.parent.type == tokens.INITIALIZES_STMT
        ):
            return NO

    elif p.type == tokens.KWARG:
        if prev.type != tokens.COMMA:
            return NO

    elif prev.type in tokens.OPENING_BRACKETS:
        return NO

    elif t == tokens.EQUAL and prev.type in tokens.ASSIGN_OPERATORS:
        return NO

    elif p.type == tokens.IMPORT:
        # handle imports
        if (
            (prev.type != tokens.IMPORT_FROM and t == tokens.DOT)
            or prev.type == tokens.DOT
            and t != tokens.IMPORT_NAME
        ):
            return NO

    elif p.type in {tokens.ATTRIBUTE, tokens.IMPORTED_TYPE}:
        # variable access
        if t == tokens.DOT or prev.type == tokens.DOT:
            return NO

    elif (
        p.type.endswith(tokens.DEF_SUFFIX)
        or p.type.endswith(tokens.GETTER_SUFIX)
        or p.type
        in {
            tokens.FUNCTION_SIG,
            tokens.CALL,
            tokens.EXTERNAL_CALL,
            tokens.EMPTY,
            tokens.SUBSCRIPT,
            tokens.INDEXED_ARGS,
            tokens.LOG_STMT,
            tokens.CONSTANT,
            tokens.IMPLEMENTS,
            tokens.USES,
            tokens.VARIABLE_ANNOTATION,
            tokens.ABI_DECODE,
        }
    ):
        # no parentheses on calls, function sigs, logs and defs
        # except for returns
        if t == tokens.LPAR and prev.type not in {tokens.RETURN_TYPE}:
            return NO

        if t == tokens.LSQB:
            return NO

    if p.type == tokens.DECORATOR:
        # decorators
        return NO

    elif t in tokens.NUMBERS and p.type in tokens.UNARY:
        # no space for signed numbers
        return NO

    return SPACE


def preceding_leaf(node: Optional[LN]) -> Optional[Leaf]:
    """Return the first leaf that precedes `node`, if any."""
    while node:
        res = node.prev_sibling
        if res:
            if isinstance(res, Leaf):
                return res

            try:
                return list(res.leaves())[-1]

            except IndexError:
                return None

        node = node.parent
    return None



import os
from pathlib import Path

import pytest
import mamushi
from mamushi.utils.output import dump_to_file

from tests.const import MINIMAL_CONTRACT, DATA_DIR
from tests.reader import read_data


def test_diff_file(runner):
    source, _ = read_data("assignments", "aug_assign_spacing.vy")
    tmp_file = Path(dump_to_file(source))
    try:
        result = runner.invoke(mamushi.main, ["--diff", str(tmp_file)])
        assert result.exit_code == 0
        assert "would reformat" in result.stderr
        assert "would be reformatted" in result.stderr

    finally:
        os.unlink(tmp_file)
    post_source, _ = read_data("assignments", "aug_assign_spacing.vy")
    assert (
        post_source == source
    ), "Source file was modified when requesting diff"


def test_diff_file_no_diffs(runner):
    tmp_file = Path(dump_to_file(MINIMAL_CONTRACT))
    try:
        result = runner.invoke(mamushi.main, ["--diff", str(tmp_file)])
        assert result.exit_code == 0
        assert "left unchanged" in result.stderr
    finally:
        os.unlink(tmp_file)


def test_diff_file_no_diffs_verbose(runner):
    tmp_file = Path(dump_to_file(MINIMAL_CONTRACT))
    try:
        result = runner.invoke(
            mamushi.main, ["--diff", "--verbose", str(tmp_file)]
        )
        assert result.exit_code == 0
        assert "already well formatted" in result.stderr
    finally:
        os.unlink(tmp_file)


def test_check_file_no_diffs(runner):
    tmp_file = Path(dump_to_file(MINIMAL_CONTRACT))
    try:
        result = runner.invoke(mamushi.main, ["--check", str(tmp_file)])
        assert result.exit_code == 0
        assert "left unchanged" in result.stderr
    finally:
        os.unlink(tmp_file)


def test_check_file(runner):
    source, _ = read_data("assignments", "aug_assign_spacing.vy")
    tmp_file = Path(dump_to_file(source))
    try:
        result = runner.invoke(mamushi.main, ["--check", str(tmp_file)])
        assert result.exit_code == 1
        assert "would reformat" in result.stderr
        assert "would be reformatted" in result.stderr

    finally:
        os.unlink(tmp_file)
    post_source, _ = read_data("assignments", "aug_assign_spacing.vy")
    assert (
        post_source == source
    ), "Source file was modified when requesting check"


def test_line_length(runner):
    source = "a: constant(uint256) = 10000000000 + 10000000000"
    tmp_file = Path(dump_to_file(source))
    try:
        result = runner.invoke(
            mamushi.main, ["--line-length", 20, str(tmp_file)]
        )
        assert result.exit_code == 0
        assert "file reformatted" in result.stderr

    finally:
        os.unlink(tmp_file)


def test_check_multiple_files(runner):
    result = runner.invoke(
        mamushi.main, ["--check", str(DATA_DIR / "imports")]
    )
    assert result.exit_code == 1
    assert "would reformat" in result.stderr
    assert "files would be reformatted" in result.stderr


def test_invalid_file(runner):
    result = runner.invoke(mamushi.main, ["--check", "AAAAAAAAAAAAAAAA.x"])
    assert result.exit_code == 2
    assert "does not exist" in result.stderr


def test_code_with_parse_error(runner):
    source = "a: (error) = invalid"
    tmp_file = Path(dump_to_file(source))
    try:
        result = runner.invoke(
            mamushi.main, ["--check", "--verbose", str(tmp_file)]
        )
        assert result.exit_code == 123
        assert "UnexpectedToken" in result.stderr
        assert "Unable to parse" in result.stderr
        assert "fail to reformat" in result.stderr

    finally:
        os.unlink(tmp_file)


def test_output_to_console(runner):
    source = "a: constant( uint256 ) = 0"
    tmp_file = Path(dump_to_file(source))
    try:
        result = runner.invoke(
            mamushi.main, ["--in-place", False, str(tmp_file)]
        )
        assert result.exit_code == 0
        assert "(uint256)" in result.stdout
        assert "file" not in result.stderr
        with open(str(tmp_file), "r") as fp:
            post_source = fp.read()
        assert (
            post_source.strip() == source.strip()
        ), "File was modified despite --in-place = False"
    finally:
        os.unlink(tmp_file)



from dataclasses import dataclass

from mamushi.__version__ import __version__
from datetime import datetime
import io
import sys
from mamushi.formatting.format import format_tree
from mamushi.parsing.comparator import compare_ast
from mamushi.parsing.parser import Parser
from mamushi.utils.files import gen_vyper_files_in_dir
import traceback
from typing import List
from pathlib import Path
import click
from mamushi.utils.output import out, diff, color_diff
from mamushi.utils.report import Report, Changed
import multiprocessing


@dataclass
class ProcessResult:
    src: Path
    success: bool
    changed: Changed | None = None
    error_message: str | None = None
    traceback_str: str | None = None
    formatted_content: str | None = None


def format_stdin_to_stdout(src: str, dst: str, file_path: Path | None):
    """Format file on stdin. Return True if changed.

    If content is None, it's read from sys.stdin.

    If `write_back` is YES, write reformatted code back to stdout. If it is DIFF,
    write a diff to stdout. The `mode` argument is passed to
    :func:`format_file_contents`.
    """
    then = datetime.utcnow()

    encoding, newline = "utf-8", ""

    f = io.TextIOWrapper(
        sys.stdout.buffer,
        encoding=encoding,
        newline=newline,
        write_through=True,
    )
    now = datetime.utcnow()
    src_name = f"STDIN\t{then} +0000 - {file_path}"
    dst_name = f"STDOUT\t{now} +0000 - {file_path}"
    d = diff(src, dst, src_name, dst_name)
    d = color_diff(d)
    f.write(d)
    f.detach()


def process_file(args):
    src, line_length, safe, diff, in_place, check = args
    parser = Parser()  # Create a new parser for each process
    return reformat(
        src=src,
        parser=parser,
        safe=safe,
        diff=diff,
        in_place=in_place and not (check or diff),
        check=check,
        line_length=line_length,
    )


def reformat(
    src: Path,
    parser: Parser,
    safe: bool,
    diff: bool,
    in_place: bool,
    check: bool,
    line_length: int,
):
    with open(src, "r") as fp:
        contract = fp.read()
    try:
        src_content = parser.parse(contract)
    except Exception:
        return ProcessResult(
            src=src,
            success=False,
            error_message="Unable to parse input file, are you sure the Vyper code is valid?",
            traceback_str=traceback.format_exc(),
        )

    res = format_tree(src_content, line_length)
    changed = Changed.NO if res == contract else Changed.YES

    if safe and not compare_ast(contract, res):
        return ProcessResult(
            src=src,
            success=False,
            error_message="Formatting changed the AST, aborting",
        )

    if check:
        return ProcessResult(src=src, success=True, changed=changed)

    if diff:
        format_stdin_to_stdout(contract, res, src)
        return ProcessResult(src=src, success=True, changed=changed)

    if in_place:
        with open(src, "w") as fp:
            fp.write(res)

    return ProcessResult(
        src=src, success=True, changed=changed, formatted_content=res
    )


@click.command()
@click.version_option(__version__)
@click.option(
    "-l",
    "--line-length",
    type=int,
    default=80,
    help="Max line length",
    show_default=True,
)
@click.option(
    "--in-place",
    type=bool,
    default=True,
    show_default=True,
    help="Overwrite files in place",
)
@click.option(
    "--safe",
    type=bool,
    default=True,
    show_default=True,
    help="Compares input and output AST to ensure similarity",
)
@click.option(
    "--check",
    is_flag=True,
    help=(
        "Don't write the files back, just return the status. Return code 0 means"
        " nothing would change. Return code 1 means some files would be reformatted."
        " Return code 123 means there was an internal error."
    ),
)
@click.option(
    "--diff",
    is_flag=True,
    help="Don't write the files back, just output a diff for each file on stdout.",
)
@click.option(
    "-q",
    "--quiet",
    is_flag=True,
    help=(
        "Don't emit non-error messages to stderr. Errors are still emitted; silence"
        " those with 2>/dev/null."
    ),
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help=(
        "Also emit messages to stderr about files that were not changed or were ignored"
        " due to exclusion patterns."
    ),
)
@click.argument(
    "src",
    nargs=-1,
    type=click.Path(exists=True, file_okay=True, dir_okay=True, readable=True),
)
@click.pass_context
def main(
    ctx: click.Context,
    line_length: int,
    in_place: bool,
    diff: bool,
    check: bool,
    verbose: bool,
    quiet: bool,
    safe: bool,
    src: List[str],
) -> None:
    sources: List[Path] = []
    if not src:
        src = [str(Path.cwd().resolve())]

    for s in src:
        p = Path(s)
        if p.is_dir():
            sources.extend(gen_vyper_files_in_dir(p))
        elif p.is_file():
            # if a file was explicitly given, we don't care about its extension
            sources.append(p)
        else:
            raise FileNotFoundError(f"invalid path: {s}")

    args_list = [
        (source, line_length, safe, diff, in_place, check)
        for source in sources
    ]

    # Use multiprocessing to process files
    with multiprocessing.Pool() as pool:
        results = pool.map(process_file, args_list)
    report = Report(check=check, diff=diff, quiet=quiet, verbose=verbose)

    for result in results:
        if not result.success:
            if verbose and result.traceback_str:
                out(result.traceback_str)
            report.failed(result.src, result.error_message)
        else:
            report.done(result.src, result.changed)
            if not in_place and result.formatted_content:
                click.echo(result.formatted_content)

    error_msg = "Oh no! 💥 💔 💥"
    if verbose or not quiet:
        out(error_msg if report.return_code else "All done! ✨ 🍰 ✨")
        if in_place:
            click.echo(str(report), err=True)
    ctx.exit(report.return_code)


if __name__ == "__main__":
    main()



import pytest
from mamushi import compare_ast


def test_comparator():
    test_input = """
# @version=0.0.1
@internal
def a():
    '''test'''
    pass
    """
    should_parse_similarly = """
# @version=0.0.1

# test

@internal
def a():
    '''docstring text does not count'''
    pass  # test
    """
    should_not_parse_similarly = """
# @version=0.0.1
@external
def b():
    pass
    """
    assert compare_ast(test_input, test_input)
    assert compare_ast(test_input, should_parse_similarly)
    assert not compare_ast(test_input, should_not_parse_similarly)



from typing import Union, List, Tuple

from mamushi.parsing import tokens
from mamushi.parsing.pytree import Leaf, Node

LN = Union[Leaf, Node]


def wrap_in_parentheses(
    parent: Node, child: LN, *, visible: bool = True
) -> None:
    """Wrap `child` in parentheses.

    This replaces `child` with an atom holding the parentheses and the old
    child.  That requires moving the prefix.

    If `visible` is False, the leaves will be valueless (and thus invisible).
    """
    lpar = Leaf(tokens.LPAR, "(" if visible else "")
    rpar = Leaf(tokens.RPAR, ")" if visible else "")
    prefix = child.prefix
    child.prefix = ""
    index = child.remove() or 0
    new_child = Node(child.type, [lpar, child, rpar])
    new_child.prefix = prefix
    parent.insert_child(index, new_child)


def is_atom_with_invisible_parens(node: LN) -> bool:
    """Given a `LN`, determines whether it's an atom `node` with invisible
    parens. Useful in dedupe-ing and normalizing parens.
    """
    if isinstance(node, Leaf) or node.type != tokens.ATOM:
        return False

    first, last = node.children[0], node.children[-1]
    return (
        isinstance(first, Leaf)
        and first.type == tokens.LPAR
        and first.value == ""
        and isinstance(last, Leaf)
        and last.type == tokens.RPAR
        and last.value == ""
    )


def ensure_visible(leaf: Leaf) -> None:
    """Make sure parentheses are visible.

    They could be invisible as part of some statements (see
    :func:`normalize_invisible_parens` and :func:`visit_import_from`).
    """
    if leaf.type == tokens.LPAR:
        leaf.value = "("
    elif leaf.type == tokens.RPAR:
        leaf.value = ")"


def is_lpar_token(nl: LN) -> bool:
    return nl.type == tokens.LPAR


def is_rpar_token(nl: LN) -> bool:
    return nl.type == tokens.RPAR


def is_one_sequence_between(
    opening: Leaf,
    closing: Leaf,
    leaves: List[Leaf],
    brackets: Tuple[str, str] = (tokens.LPAR, tokens.RPAR),
) -> bool:
    """Return True if content between `opening` and `closing` is a one-sequence."""
    if (opening.type, closing.type) != brackets:
        return False

    depth = closing.bracket_depth + 1
    for _opening_index, leaf in enumerate(leaves):
        if leaf is opening:
            break

    else:
        raise LookupError("Opening paren not found in `leaves`")

    commas = 0
    _opening_index += 1
    for leaf in leaves[_opening_index:]:
        if leaf is closing:
            break

        bracket_depth = leaf.bracket_depth
        if bracket_depth == depth and leaf.type == tokens.COMMA:
            commas += 1
            if leaf.parent and leaf.parent.type in {
                tokens.ARGUMENTS,
                tokens.PARAMETERS,
            }:
                commas += 1
                break

    return commas < 2


def replace_child(old_child: LN, new_child: LN) -> None:
    """
    Side Effects:
        * If @old_child.parent is set, replace @old_child with @new_child in
        @old_child's underlying Node structure.
            OR
        * Otherwise, this function does nothing.
    """
    parent = old_child.parent
    if not parent:
        return

    child_idx = old_child.remove()
    if child_idx is not None:
        parent.insert_child(child_idx, new_child)



from tests.reader import all_data_cases, read_data, all_data
import pytest
from mamushi.formatting.format import format_tree

test_cases = [
    (category, case)
    for category in all_data()
    for case in all_data_cases(category)
]


@pytest.mark.parametrize("category,case", test_cases)
def test_format(case: str, category: str, parser):
    source, expected = read_data(category, case)
    assert format_tree(parser.parse(source)).strip() == expected.strip()



import itertools
import sys
from dataclasses import dataclass, field
from functools import wraps
from typing import (
    List,
    Optional,
    Collection,
    Iterator,
    Callable,
    Sequence,
    Tuple,
    TypeVar,
    Dict,
    cast,
    Set,
)
from mamushi.formatting.whitespace import whitespace
from mamushi.formatting.brackets import (
    BracketTracker,
    COMMA_PRIORITY,
    DOT_PRIORITY,
)
from mamushi.formatting.nodes import (
    ensure_visible,
    is_one_sequence_between,
    replace_child,
)
from mamushi.parsing import tokens
from mamushi.parsing.pytree import Leaf, Node


LeafID = int
T = TypeVar("T")
Index = int


class CannotTransform(Exception):
    """Base class for errors raised by Transformers."""


class CannotSplit(CannotTransform):
    """A readable split that fits the allotted line length is impossible."""


@dataclass
class Line:
    """Holds leaves and comments. Can be printed with `str(line)`."""

    depth: int = 0
    leaves: List[Leaf] = field(default_factory=list)
    # keys ordered like `leaves`
    comments: Dict[LeafID, List[Leaf]] = field(default_factory=dict)
    bracket_tracker: BracketTracker = field(default_factory=BracketTracker)
    inside_brackets: bool = False
    should_split_rhs: bool = False
    magic_trailing_comma: Optional[Leaf] = None

    def append(self, leaf: Leaf, preformatted: bool = False) -> None:
        """Add a new `leaf` to the end of the line.

        Unless `preformatted` is True, the `leaf` will receive a new consistent
        whitespace prefix and metadata applied by :class:`BracketTracker`.
        Trailing commas are maybe removed, unpacked for loop variables are
        demoted from being delimiters.

        Inline comments are put aside.
        """
        has_value = leaf.type in tokens.BRACKETS or bool(leaf.value.strip())
        if not has_value:
            return

        if self.leaves and not preformatted:
            # Note: at this point leaf.prefix should be empty except for
            # imports, for which we only preserve newlines.

            leaf.prefix += whitespace(leaf)
        if self.inside_brackets or not preformatted:
            self.bracket_tracker.mark(leaf)
            if self.has_magic_trailing_comma(leaf):
                self.magic_trailing_comma = leaf
        if not self.append_comment(leaf):
            self.leaves.append(leaf)

    def append_safe(self, leaf: Leaf, preformatted: bool = False) -> None:
        """Like :func:`append()` but disallow invalid standalone comment structure.

        Raises ValueError when any `leaf` is appended after a standalone comment
        or when a standalone comment is not the first leaf on the line.
        """
        if self.bracket_tracker.depth == 0:
            if self.is_comment:
                raise ValueError("cannot append to standalone comments")

            if self.leaves and leaf.type == tokens.STANDALONE_COMMENT:
                raise ValueError(
                    "cannot append standalone comments to a populated line"
                )

        self.append(leaf, preformatted=preformatted)

    @property
    def is_comment(self) -> bool:
        """Is this line a standalone comment?"""
        return len(self.leaves) == 1 and self.leaves[0].type in {
            tokens.STANDALONE_COMMENT,
            tokens.DOCSTRING,
        }

    @property
    def is_decorator(self) -> bool:
        """Is this line a decorator?"""
        return bool(self) and self.leaves[0].type == tokens.AT

    @property
    def is_import(self) -> bool:
        """Is this an import line?"""
        return bool(self) and is_import(self.leaves[0])

    @property
    def is_flow_control(self) -> bool:
        return (
            bool(self)
            and self.leaves[0].type == tokens.NAME
            and self.leaves[0].value in tokens.FLOW_CONTROL
        )

    @property
    def is_pragma(self) -> bool:
        try:
            first_leaf = self.leaves[0]
        except IndexError:
            return False
        return first_leaf.type == tokens.PRAGMA

    @property
    def is_def(self) -> bool:
        try:
            first_leaf = self.leaves[0]
        except IndexError:
            return False

        is_abi = (
            first_leaf.parent
            and first_leaf.parent.parent
            and first_leaf.parent.parent.type == tokens.INTERFACE_FUNCTION
        )
        return (first_leaf.type in tokens.DECLARATIONS) and not is_abi

    def contains_standalone_comments(
        self, depth_limit: int = sys.maxsize
    ) -> bool:
        """If so, needs to be split before emitting."""
        for leaf in self.leaves:
            if (
                leaf.type == tokens.STANDALONE_COMMENT
                and leaf.bracket_depth <= depth_limit
            ):
                return True

        return False

    def has_magic_trailing_comma(self, closing: Leaf) -> bool:
        """Return True if we have a magic trailing comma, that is when:
        - there's a trailing comma here
        - it's not a one-tuple
        - it's not a single-element subscript
        Additionally, if ensure_removable:
        - it's not from square bracket indexing
        (specifically, single-element square bracket indexing with
        Preview.skip_magic_trailing_comma_in_subscript)
        """
        if not (
            closing.type in tokens.CLOSING_BRACKETS
            and self.leaves
            and self.leaves[-1].type == tokens.COMMA
        ):
            return False

        if closing.type == tokens.RBRACE:
            return True

        if closing.type == tokens.RSQB:
            return True

        if self.is_import:
            return True

        if (
            closing.opening_bracket is not None
            and not is_one_sequence_between(
                closing.opening_bracket, closing, self.leaves
            )
        ):
            return True

        return False

    def append_comment(self, comment: Leaf) -> bool:
        """Add an inline or standalone comment to the line."""
        if (
            comment.type == tokens.STANDALONE_COMMENT
            and self.bracket_tracker.any_open_brackets()
        ):
            comment.prefix = ""
            return False

        if comment.type != tokens.COMMENT:
            return False

        if not self.leaves:
            comment.type = tokens.STANDALONE_COMMENT
            comment.prefix = ""
            return False

        last_leaf = self.leaves[-1]
        if (
            last_leaf.type == tokens.RPAR
            and not last_leaf.value
            and last_leaf.parent
            and len(list(last_leaf.parent.leaves())) <= 3
        ):
            # Comments on an optional parens wrapping a single leaf should belong to
            # the wrapped node except if it's a type comment. Pinning the comment like
            # this avoids unstable formatting caused by comment migration.
            if len(self.leaves) < 2:
                comment.type = tokens.STANDALONE_COMMENT
                comment.prefix = ""
                return False

            last_leaf = self.leaves[-2]
        self.comments.setdefault(id(last_leaf), []).append(comment)
        return True

    def comments_after(self, leaf: Leaf) -> List[Leaf]:
        """Generate comments that should appear directly after `leaf`."""
        return self.comments.get(id(leaf), [])

    def enumerate_with_length(
        self, reversed: bool = False
    ) -> Iterator[Tuple[Index, Leaf, int]]:
        """Return an enumeration of leaves with their length.

        Stops prematurely on standalone comments.
        """
        op = cast(
            Callable[[Sequence[Leaf]], Iterator[Tuple[Index, Leaf]]],
            enumerate_reversed if reversed else enumerate,
        )
        for index, leaf in op(self.leaves):
            length = len(leaf.prefix) + len(leaf.value)

            for comment in self.comments_after(leaf):
                length += len(comment.value)

            yield index, leaf, length

    def clone(self) -> "Line":
        return Line(
            depth=self.depth,
            inside_brackets=self.inside_brackets,
            should_split_rhs=self.should_split_rhs,
            magic_trailing_comma=self.magic_trailing_comma,
        )

    def __str__(self) -> str:
        """Render the line."""
        if not self:
            return "\n"

        indent = "    " * self.depth
        leaves = iter(self.leaves)
        first = next(leaves)
        res = f"{first.prefix}{indent}{first.value}"
        for leaf in leaves:
            res += str(leaf)
        for comment in itertools.chain.from_iterable(self.comments.values()):
            res += str(comment)

        return res + "\n"

    def __bool__(self) -> bool:
        """Return True if the line has leaves or comments."""
        return bool(self.leaves or self.comments)


def right_hand_split(
    line: Line,
    line_length: int,
    omit: Collection[LeafID] = (),
    force_optional_parentheses: bool = False,
) -> Iterator[Line]:
    """Split line into many lines, starting with the last matching bracket pair.

    If the split was by optional parentheses, attempt splitting without them, too.
    `omit` is a collection of closing bracket IDs that shouldn't be considered for
    this split.

    Note: running this function modifies `bracket_depth` on the leaves of `line`.
    """
    tail_leaves: List[Leaf] = []
    body_leaves: List[Leaf] = []
    head_leaves: List[Leaf] = []
    current_leaves = tail_leaves
    opening_bracket: Optional[Leaf] = None
    closing_bracket: Optional[Leaf] = None
    for leaf in reversed(line.leaves):
        if current_leaves is body_leaves:
            if leaf is opening_bracket:
                current_leaves = head_leaves if body_leaves else tail_leaves
        current_leaves.append(leaf)
        if current_leaves is tail_leaves:
            if leaf.type in tokens.CLOSING_BRACKETS and id(leaf) not in omit:
                opening_bracket = leaf.opening_bracket
                closing_bracket = leaf
                current_leaves = body_leaves
    if not (opening_bracket and closing_bracket and head_leaves):
        # If there is no opening or closing_bracket that means the split failed and
        # all content is in the tail.  Otherwise, if `head_leaves` are empty, it means
        # the matching `opening_bracket` wasn't available on `line` anymore.
        raise CannotSplit("No brackets found")

    tail_leaves.reverse()
    body_leaves.reverse()
    head_leaves.reverse()
    head = bracket_split_build_line(head_leaves, line, opening_bracket)
    body = bracket_split_build_line(
        body_leaves, line, opening_bracket, is_body=True
    )
    tail = bracket_split_build_line(tail_leaves, line, opening_bracket)
    bracket_split_succeeded_or_raise(head, body, tail)
    if (
        not force_optional_parentheses
        # the opening bracket is an optional paren
        and opening_bracket.type == tokens.LPAR
        and not opening_bracket.value
        # the closing bracket is an optional paren
        and closing_bracket.type == tokens.RPAR
        and not closing_bracket.value
        # it's not an import (optional parens are the only thing we can split on
        # in this case; attempting a split without them is a waste of time)
        and not line.is_import
        # there are no standalone comments in the body
        and not body.contains_standalone_comments(0)
        # and we can actually remove the parens
        and can_omit_invisible_parens(body, line_length)
    ):
        omit = {id(closing_bracket), *omit}
        try:
            yield from right_hand_split(line, line_length, omit=omit)
            return

        except CannotSplit as e:
            if not (is_line_short_enough(body, line_length=line_length)):
                raise CannotSplit(
                    "Splitting failed, body is still too long and can't be split."
                ) from e

    ensure_visible(opening_bracket)
    ensure_visible(closing_bracket)
    for result in (head, body, tail):
        if result:
            yield result


def bracket_split_build_line(
    leaves: List[Leaf],
    original: Line,
    opening_bracket: Leaf,
    *,
    is_body: bool = False,
) -> Line:
    """Return a new line with given `leaves` and respective comments from `original`.

    If `is_body` is True, the result line is one-indented inside brackets and as such
    has its first leaf's prefix normalized and a trailing comma added when expected.
    """
    result = Line(depth=original.depth)
    if is_body:
        result.inside_brackets = True
        result.depth += 1
        if leaves:
            # Ensure a trailing comma for imports and standalone function arguments, but
            # be careful not to add one after any comments or within type annotations.
            no_commas = (
                original.is_def
                and opening_bracket.value == "("
                and not any(leaf.type == tokens.COMMA for leaf in leaves)
                # In particular, don't add one within a parenthesized return annotation.
                # Unfortunately the indicator we're in a return annotation (RARROW) may
                # be defined directly in the parent node, the parent of the parent ...
                # and so on depending on how complex the return annotation is.
                # This isn't perfect and there's some false negatives but they are in
                # contexts were a comma is actually fine.
                and not any(
                    node.prev_sibling.type == tokens.RETURN_TYPE
                    for node in (
                        leaves[0].parent,
                        getattr(leaves[0].parent, "parent", None),
                    )
                    if isinstance(node, Node)
                    and isinstance(node.prev_sibling, Leaf)
                )
            )

            if original.is_import or no_commas:
                for i in range(len(leaves) - 1, -1, -1):
                    if leaves[i].type == tokens.STANDALONE_COMMENT:
                        continue

                    if leaves[i].type != tokens.COMMA:
                        new_comma = Leaf(tokens.COMMA, ",")
                        leaves.insert(i + 1, new_comma)
                    break

    # Populate the line
    for leaf in leaves:
        result.append(leaf, preformatted=True)
        for comment_after in original.comments_after(leaf):
            result.append(comment_after, preformatted=True)
    if is_body and should_split_line(result, opening_bracket):
        result.should_split_rhs = True
    return result


def should_split_line(line: Line, opening_bracket: Leaf) -> bool:
    """Should `line` be immediately split with `delimiter_split()` after RHS?"""

    if not (opening_bracket.parent and opening_bracket.value in "[{("):
        return False

    # We're essentially checking if the body is delimited by commas and there's more
    # than one of them (we're excluding the trailing comma and if the delimiter priority
    # is still commas, that means there's more).
    exclude = set()
    trailing_comma = False
    try:
        last_leaf = line.leaves[-1]
        if last_leaf.type == tokens.COMMA:
            trailing_comma = True
            exclude.add(id(last_leaf))
        max_priority = line.bracket_tracker.max_delimiter_priority(
            exclude=exclude
        )
    except (IndexError, ValueError):
        return False

    return max_priority == COMMA_PRIORITY and (
        trailing_comma
        # always explode imports
        # TODO: check for imports / mid-level ATOM
        or opening_bracket.parent.type in {tokens.ATOM}
    )


def line_to_string(line: Line) -> str:
    """Returns the string representation of @line.

    WARNING: This is known to be computationally expensive.
    """
    return str(line).strip("\n")


def is_line_short_enough(
    line: Line, line_length: int, line_str: str = ""
) -> bool:
    """Return True if `line` is no longer than `line_length`.

    Uses the provided `line_str` rendering, if any, otherwise computes a new one.
    """
    if not line_str:
        line_str = line_to_string(line)
    return (
        len(line_str) <= line_length
        and "\n" not in line_str  # multiline strings
        and not line.contains_standalone_comments()
    )


def can_omit_invisible_parens(
    line: Line,
    line_length: int,
) -> bool:
    """Does `line` have a shape safe to reformat without optional parens around it?

    Returns True for only a subset of potentially nice looking formattings but
    the point is to not return false positives that end up producing lines that
    are too long.
    """
    bt = line.bracket_tracker
    if not bt.delimiters:
        # Without delimiters the optional parentheses are useless.
        return True

    max_priority = bt.max_delimiter_priority()
    if bt.delimiter_count_with_priority(max_priority) > 1:
        # With more than one delimiter of a kind the optional parentheses read better.
        return False

    if max_priority == DOT_PRIORITY:
        # A single stranded method call doesn't require optional parentheses.
        return True

    assert len(line.leaves) >= 2, "Stranded delimiter"

    # With a single delimiter, omit if the expression starts or ends with
    # a bracket.
    first = line.leaves[0]
    second = line.leaves[1]
    if (
        first.type in tokens.OPENING_BRACKETS
        and second.type not in tokens.CLOSING_BRACKETS
    ):
        if _can_omit_opening_paren(line, first=first, line_length=line_length):
            return True

        # Note: we are not returning False here because a line might have *both*
        # a leading opening bracket and a trailing closing bracket.  If the
        # opening bracket doesn't match our rule, maybe the closing will.

    penultimate = line.leaves[-2]
    last = line.leaves[-1]

    if last.type == tokens.RPAR or last.type == tokens.RBRACE:
        if penultimate.type in tokens.OPENING_BRACKETS:
            # Empty brackets don't help.
            return False

        if _can_omit_closing_paren(line, last=last, line_length=line_length):
            return True

    return False


def _can_omit_opening_paren(
    line: Line, *, first: Leaf, line_length: int
) -> bool:
    """See `can_omit_invisible_parens`."""
    remainder = False
    length = 4 * line.depth
    _index = -1
    for _index, leaf, leaf_length in line.enumerate_with_length():
        if (
            leaf.type in tokens.CLOSING_BRACKETS
            and leaf.opening_bracket is first
        ):
            remainder = True
        if remainder:
            length += leaf_length
            if length > line_length:
                break

            if leaf.type in tokens.OPENING_BRACKETS:
                # There are brackets we can further split on.
                remainder = False

    else:
        # checked the entire string and line length wasn't exceeded
        if len(line.leaves) == _index + 1:
            return True

    return False


def _can_omit_closing_paren(
    line: Line, *, last: Leaf, line_length: int
) -> bool:
    """See `can_omit_invisible_parens`."""
    length = 4 * line.depth
    seen_other_brackets = False
    for _index, leaf, leaf_length in line.enumerate_with_length():
        length += leaf_length
        if leaf is last.opening_bracket:
            if seen_other_brackets or length <= line_length:
                return True

        elif leaf.type in tokens.OPENING_BRACKETS:
            # There are brackets we can further split on.
            seen_other_brackets = True

    return False


def bracket_split_succeeded_or_raise(
    head: Line, body: Line, tail: Line
) -> None:
    """Raise :exc:`CannotSplit` if the last left- or right-hand split failed.

    Do nothing otherwise.

    A left- or right-hand split is based on a pair of brackets. Content before
    (and including) the opening bracket is left on one line, content inside the
    brackets is put on a separate line, and finally content starting with and
    following the closing bracket is put on a separate line.

    Those are called `head`, `body`, and `tail`, respectively. If the split
    produced the same line (all content in `head`) or ended up with an empty `body`
    and the `tail` is just the closing bracket, then it's considered failed.
    """
    tail_len = len(str(tail).strip())
    if not body:
        if tail_len == 0:
            raise CannotSplit("Splitting brackets produced the same line")

        elif tail_len < 3:
            raise CannotSplit(
                f"Splitting brackets on an empty body to save {tail_len} characters is"
                " not worth it"
            )


def is_import(leaf: Leaf) -> bool:
    """Returns True if the given leaf starts an import statement."""
    p = leaf.parent
    t = leaf.type
    v = leaf.value
    return bool(
        t in {tokens.IMPORT_FROM, tokens.IMPORT}
        and (
            (v == "import" and p and p.type == "import")
            or (v == "from" and p and p.type == "import")
        )
    )


def enumerate_reversed(sequence: Sequence[T]) -> Iterator[Tuple[Index, T]]:
    """Like `reversed(enumerate(sequence))` if that were possible."""
    index = len(sequence) - 1
    for element in reversed(sequence):
        yield (index, element)
        index -= 1


def left_hand_split(
    line: Line, force_optional_parentheses: bool = False
) -> Iterator[Line]:
    """Split line into many lines, starting with the first matching bracket pair.

    Note: this usually looks weird, only use this for function definitions.
    Prefer RHS otherwise.  This is why this function is not symmetrical with
    :func:`right_hand_split` which also handles optional parentheses.
    """
    tail_leaves: List[Leaf] = []
    body_leaves: List[Leaf] = []
    head_leaves: List[Leaf] = []
    current_leaves = head_leaves
    matching_bracket: Optional[Leaf] = None
    for leaf in line.leaves:
        if (
            current_leaves is body_leaves
            and leaf.type in tokens.CLOSING_BRACKETS
            and leaf.opening_bracket is matching_bracket
            and isinstance(matching_bracket, Leaf)
        ):
            ensure_visible(leaf)
            ensure_visible(matching_bracket)
            current_leaves = tail_leaves if body_leaves else head_leaves
        current_leaves.append(leaf)
        if current_leaves is head_leaves:
            if leaf.type in tokens.OPENING_BRACKETS:
                matching_bracket = leaf
                current_leaves = body_leaves
    if not matching_bracket:
        raise CannotSplit("No brackets found")

    head = bracket_split_build_line(head_leaves, line, matching_bracket)
    body = bracket_split_build_line(
        body_leaves, line, matching_bracket, is_body=True
    )
    tail = bracket_split_build_line(tail_leaves, line, matching_bracket)
    bracket_split_succeeded_or_raise(head, body, tail)
    for result in (head, body, tail):
        if result:
            yield result


def normalize_prefix(leaf: Leaf, *, inside_brackets: bool) -> None:
    """Leave existing extra newlines if not `inside_brackets`. Remove everything
    else.

    Note: don't use backslashes for formatting or you'll lose your voting rights.
    """
    if not inside_brackets:
        spl = leaf.prefix.split("#")
        if "\\" not in spl[0]:
            nl_count = spl[-1].count("\n")
            if len(spl) > 1:
                nl_count -= 1
            leaf.prefix = "\n" * nl_count
            return

    leaf.prefix = ""


def dont_increase_indentation(split_func: Callable):
    """Normalize prefix of the first leaf in every line returned by `split_func`.

    This is a decorator over relevant split functions.
    """

    @wraps(split_func)
    def split_wrapper(
        line: Line, force_optional_parentheses: bool = False
    ) -> Iterator[Line]:
        for split_line in split_func(line):
            normalize_prefix(split_line.leaves[0], inside_brackets=True)
            yield split_line

    return split_wrapper


@dont_increase_indentation
def delimiter_split(
    line: Line, force_optional_parentheses: bool = False
) -> Iterator[Line]:
    """Split according to delimiters of the highest priority.

    If the appropriate Features are given, the split will add trailing commas
    also in function signatures and calls that contain `*` and `**`.
    """
    try:
        last_leaf = line.leaves[-1]
    except IndexError:
        raise CannotSplit("Line empty") from None

    bt = line.bracket_tracker
    try:
        delimiter_priority = bt.max_delimiter_priority(exclude={id(last_leaf)})
    except ValueError:
        raise CannotSplit("No delimiters found") from None

    if delimiter_priority == DOT_PRIORITY:
        if bt.delimiter_count_with_priority(delimiter_priority) == 1:
            raise CannotSplit(
                "Splitting a single attribute from its owner looks wrong"
            )

    current_line = Line(depth=line.depth, inside_brackets=line.inside_brackets)

    def append_to_line(leaf: Leaf) -> Iterator[Line]:
        """Append `leaf` to current line or to new line if appending impossible."""
        nonlocal current_line
        try:
            current_line.append_safe(leaf, preformatted=True)
        except ValueError:
            yield current_line

            current_line = Line(
                depth=line.depth, inside_brackets=line.inside_brackets
            )
            current_line.append(leaf)

    for leaf in line.leaves:
        yield from append_to_line(leaf)

        for comment_after in line.comments_after(leaf):
            yield from append_to_line(comment_after)

        leaf_priority = bt.delimiters.get(id(leaf))
        if leaf_priority == delimiter_priority:
            yield current_line

            current_line = Line(
                depth=line.depth, inside_brackets=line.inside_brackets
            )
    if current_line:
        if (
            delimiter_priority == COMMA_PRIORITY
            and current_line.leaves[-1].type != tokens.COMMA
            and current_line.leaves[-1].type != tokens.STANDALONE_COMMENT
        ):
            new_comma = Leaf(tokens.COMMA, ",")
            current_line.append(new_comma)
        yield current_line


def split_line(
    line: Line,
    line_length: int,
    inner: bool = False,
    force_optional_parentheses: bool = False,
) -> Iterator[Line]:
    """Splits a `line` into potentially many lines.
    They should fit in the allotted `line_length` but might not be able to.
    `inner` signifies that there were a pair of brackets somewhere around the
    current `line`, possibly transitively. This means we can fallback to splitting
    by delimiters if the LHS/RHS don't yield any results.
    """
    if line.is_comment:
        yield line
        return

    line_str = str(line).strip("\n")
    if (
        not line.should_split_rhs
        and not line.magic_trailing_comma
        and (is_line_short_enough(line, line_length, line_str=line_str))
        and not (line.inside_brackets and line.contains_standalone_comments())
    ):
        transformers = []

    elif line.is_def:
        transformers = [left_hand_split]
    else:

        def _rhs(
            self: object, line: Line, force_optional_parentheses=False
        ) -> Iterator[Line]:
            """Wraps calls to `right_hand_split`.

            The calls increasingly `omit` right-hand trailers (bracket pairs with
            content), meaning the trailers get glued together to split on another
            bracket pair instead.
            """
            for omit in generate_trailers_to_omit(line, line_length):
                lines = list(
                    right_hand_split(
                        line,
                        line_length,
                        omit=omit,
                        force_optional_parentheses=force_optional_parentheses,
                    )
                )
                # Note: this check is only able to figure out if the first line of the
                # *current* transformation fits in the line length.  This is true only
                # for simple cases.  All others require running more transforms via
                # `transform_line()`.  This check doesn't know if those would succeed.
                if is_line_short_enough(lines[0], line_length=line_length):
                    yield from lines
                    return

            # All splits failed, best effort split with no omits.
            # This mostly happens to multiline strings that are by definition
            # reported as not fitting a single line, as well as lines that contain
            # trailing commas (those have to be exploded).
            yield from right_hand_split(
                line,
                line_length=line_length,
                force_optional_parentheses=force_optional_parentheses,
            )

        # HACK: nested functions (like _rhs) compiled by mypyc don't retain their
        # __name__ attribute which is needed in `run_transformer` further down.
        # Unfortunately a nested class breaks mypyc too. So a class must be created
        # via type ... https://github.com/mypyc/mypyc/issues/884
        rhs = type("rhs", (), {"__call__": _rhs})()

        if line.inside_brackets:
            transformers = [delimiter_split, rhs]
        else:
            transformers = [rhs]

    transformers.append(hug_power_op)

    for transform in transformers:
        # We are accumulating lines in `result` because we might want to abort
        # mission and return the original line in the end, or attempt a different
        # split altogether.
        try:
            result = run_transformer(
                line,
                transform,
                line_length=line_length,
                line_str=line_str,
                force_optional_parentheses=force_optional_parentheses,
            )
        except CannotTransform:
            continue
        else:
            yield from result
            break

    else:
        yield line


def generate_trailers_to_omit(
    line: Line, line_length: int
) -> Iterator[Set[LeafID]]:
    """Generate sets of closing bracket IDs that should be omitted in a RHS.

    Brackets can be omitted if the entire trailer up to and including
    a preceding closing bracket fits in one line.

    Yielded sets are cumulative (contain results of previous yields, too).  First
    set is empty, unless the line should explode, in which case bracket pairs until
    the one that needs to explode are omitted.
    """

    omit: Set[LeafID] = set()
    if not line.magic_trailing_comma:
        yield omit

    length = 4 * line.depth
    opening_bracket: Optional[Leaf] = None
    closing_bracket: Optional[Leaf] = None
    inner_brackets: Set[LeafID] = set()
    for index, leaf, leaf_length in line.enumerate_with_length(reversed=True):
        length += leaf_length
        if length > line_length:
            break

        has_inline_comment = leaf_length > len(leaf.value) + len(leaf.prefix)
        if leaf.type == tokens.STANDALONE_COMMENT or has_inline_comment:
            break

        if opening_bracket:
            if leaf is opening_bracket:
                opening_bracket = None
            elif leaf.type in tokens.CLOSING_BRACKETS:
                prev = line.leaves[index - 1] if index > 0 else None
                if (
                    prev
                    and prev.type == tokens.COMMA
                    and leaf.opening_bracket is not None
                    and not is_one_sequence_between(
                        leaf.opening_bracket, leaf, line.leaves
                    )
                ):
                    # Never omit bracket pairs with trailing commas.
                    # We need to explode on those.
                    break

                inner_brackets.add(id(leaf))
        elif leaf.type in tokens.CLOSING_BRACKETS:
            prev = line.leaves[index - 1] if index > 0 else None
            if prev and prev.type in tokens.OPENING_BRACKETS:
                # Empty brackets would fail a split so treat them as "inner"
                # brackets (e.g. only add them to the `omit` set if another
                # pair of brackets was good enough.
                inner_brackets.add(id(leaf))
                continue

            if closing_bracket:
                omit.add(id(closing_bracket))
                omit.update(inner_brackets)
                inner_brackets.clear()
                yield omit

            if (
                prev
                and prev.type == tokens.COMMA
                and leaf.opening_bracket is not None
                and not is_one_sequence_between(
                    leaf.opening_bracket, leaf, line.leaves
                )
            ):
                # Never omit bracket pairs with trailing commas.
                # We need to explode on those.
                break

            if leaf.value:
                opening_bracket = leaf.opening_bracket
                closing_bracket = leaf


def run_transformer(
    line: Line,
    transform: Callable[[Line, bool], Iterator[Line]],
    line_length: int = 80,
    line_str: str = "",
    force_optional_parentheses=False,
) -> List[Line]:
    if not line_str:
        line_str = line_to_string(line)
    result: List[Line] = []
    for transformed_line in transform(line, force_optional_parentheses):
        if str(transformed_line).strip("\n") == line_str:
            raise CannotTransform(
                "Line transformer returned an unchanged result"
            )

        result.extend(
            split_line(
                transformed_line,
                line_length,
                force_optional_parentheses=force_optional_parentheses,
            )
        )

    if (
        transform.__class__.__name__ != "rhs"
        or not line.bracket_tracker.invisible
        or any(bracket.value for bracket in line.bracket_tracker.invisible)
        or is_line_short_enough(result[0], line_length=line_length)
        # If any leaves have no parents (which _can_ occur since
        # `transform(line)` potentially destroys the line's underlying node
        # structure), then we can't proceed. Doing so would cause the below
        # call to `append_leaves()` to fail.
        or any(leaf.parent is None for leaf in line.leaves)
    ):
        return result

    line_copy = line.clone()
    append_leaves(line_copy, line, line.leaves)
    second_opinion = run_transformer(
        line_copy,
        transform,
        line_str=line_str,
        force_optional_parentheses=True,
    )
    if all(
        is_line_short_enough(ln, line_length=line_length)
        for ln in second_opinion
    ):
        result = second_opinion
    return result


def append_leaves(
    new_line: Line,
    old_line: Line,
    leaves: List[Leaf],
    preformatted: bool = False,
) -> None:
    """
    Append leaves (taken from @old_line) to @new_line, making sure to fix the
    underlying Node structure where appropriate.

    All of the leaves in @leaves are duplicated. The duplicates are then
    appended to @new_line and used to replace their originals in the underlying
    Node structure. Any comments attached to the old leaves are reattached to
    the new leaves.

    Pre-conditions:
        set(@leaves) is a subset of set(@old_line.leaves).
    """
    for old_leaf in leaves:
        new_leaf = Leaf(old_leaf.type, old_leaf.value)
        replace_child(old_leaf, new_leaf)
        new_line.append(new_leaf, preformatted=preformatted)

        for comment_leaf in old_line.comments_after(old_leaf):
            new_line.append(comment_leaf, preformatted=True)


def hug_power_op(
    line: Line, force_optional_parentheses: bool = False
) -> Iterator[Line]:
    """A transformer which normalizes spacing around power operators."""

    # Performance optimization to avoid unnecessary Leaf clones and other ops.
    for leaf in line.leaves:
        if leaf.type == tokens.DOUBLESTAR:
            break
    else:
        raise CannotTransform("No doublestar token was found in the line.")

    new_line = line.clone()
    should_hug = False
    for idx, leaf in enumerate(line.leaves):
        new_leaf = leaf.clone()
        if should_hug:
            new_leaf.prefix = ""
            should_hug = False

        should_hug = (0 < idx < len(line.leaves) - 1) and (
            leaf.type == tokens.DOUBLESTAR
            # we don't want to hug if we're in x **= y
            and bool(
                leaf.parent
                and leaf.parent.parent
                and leaf.parent.parent.type != tokens.AUG_ASSIGN
            )
        )
        if should_hug:
            new_leaf.prefix = ""

        # We have to be careful to make a new line properly:
        # - bracket related metadata must be maintained (handled by Line.append)
        # - comments need to copied over, updating the leaf IDs they're attached to
        new_line.append(new_leaf, preformatted=True)
        for comment_leaf in line.comments_after(leaf):
            new_line.append(comment_leaf, preformatted=True)

    yield new_line



import re

from mamushi.parsing.pytree import Leaf


def add_leading_space_after_hashtag(comment: str) -> str:
    """Add a space between hashtag and first character if none are there"""
    return re.sub(r"(^#)(\w)", r"# \2", comment)


def settle_prefix(comment: Leaf):
    """Puts the leading line returns in prefix"""
    stripped = comment.value.lstrip("\n")
    comment.prefix += "\n" * (len(comment.value) - len(stripped))
    comment.value = stripped



from lark import Token
from lark.visitors import Transformer_InPlaceRecursive

from mamushi.parsing import tokens
from lark import Lark, Tree
from mamushi.parsing.parser import PythonIndenter

_plain_lark_grammar = None


class NullifyStringsAndNewLines(Transformer_InPlaceRecursive):
    def DOCSTRING(self, args):
        return Token(type=tokens.DOCSTRING, value="")

    def _NEWLINE(self, args):
        return Token(type=tokens.NEWLINE, value="")

    def STRING(self, args):
        return Token(type=tokens.STRING, value="")


def plain_grammar():
    global _plain_lark_grammar
    if _plain_lark_grammar is None:
        _plain_lark_grammar = Lark.open_from_package(
            "mamushi",
            "grammar.lark",
            ["parsing"],
            parser="lalr",
            start="module",
            postlex=PythonIndenter(),
            keep_all_tokens=False,
            maybe_placeholders=False,
        )
    return _plain_lark_grammar


def parse_string_to_tokenless_ast(code: str) -> Tree:
    return plain_grammar().parse(code + "\n\n")


def compare_ast(src: str, dest: str) -> bool:
    """
    Generates two ASTs with no punctuation tokens  and uses a transformer to
    remove docstrings (since these are reformatted and not ignored)
    and newlines (since we might reformat comments) then compares
    the two trees to ensure no syntactic changes were introduced
    """
    src_tree = parse_string_to_tokenless_ast(src)
    dest_tree = parse_string_to_tokenless_ast(dest)
    NullifyStringsAndNewLines().transform(src_tree)
    NullifyStringsAndNewLines().transform(dest_tree)
    return src_tree == dest_tree



# EXPERIMENTAL VYPER PARSER
# https://github.com/vyperlang/vyper/
import re
from collections import defaultdict
from typing import Any, Dict, List, Tuple

from lark import Lark, Token, Tree

from mamushi.parsing import tokens
from mamushi.parsing.indenter import PythonIndenter
from mamushi.parsing.pytree import Leaf, Node
from mamushi.parsing.tokens import NEWLINE

CommentMapping = Dict[Tuple[Any, ...], Token]
StandAloneCommentMapping = Dict[Tuple[Any, ...], List[Token]]
DedentCount = Dict[Tuple[Any, ...], int]


class Parser(object):
    """
    Parse a vyper file using lark, converts the lark tree to a pytree
    Lalr needs comments that aren't parsed along with newlines to be ignored
    We collect those with callbacks and then reintegrate them during the
    conversion to pytree
    """

    def __init__(self):
        self._comments = []
        self._all_newlines = []
        self._comment_mapping: CommentMapping = {}
        self._orphan_comment_mapping: StandAloneCommentMapping = defaultdict(
            list
        )
        self._header_comments: List[Token] = []
        self.indenter = PythonIndenter()
        self.lalr = self._create_lalr_parser()

    def _create_lalr_parser(self):
        return Lark.open_from_package(
            "mamushi",
            "grammar.lark",
            ["parsing"],
            parser="lalr",
            start="module",
            postlex=self.indenter,
            keep_all_tokens=True,
            propagate_positions=True,
            lexer_callbacks={
                "COMMENT": self._comments.append,
                "_NEWLINE": self._record_all_newlines,
            },
            maybe_placeholders=False,
        )

    def _record_all_newlines(self, token: Token):
        self._all_newlines.append(token)
        return token

    @staticmethod
    def _preprocess(code: str) -> str:
        return re.sub(r"[ \t]+\n", "\n", code, flags=re.MULTILINE) + "\n"

    def _clear_comments(self):
        self._comments.clear()
        self._comment_mapping.clear()
        self._orphan_comment_mapping.clear()
        self._header_comments.clear()
        self._all_newlines.clear()

    def parse(self, code):
        self._clear_comments()
        lark_tree = self.lalr.parse(self._preprocess(code))
        self._generate_comment_associations(lark_tree)
        pytree = self._to_pytree(lark_tree)
        return pytree

    @staticmethod
    def _break_down_comments(t: Token) -> List[Token]:
        res = []
        nlines = 0
        newline = t.value.removeprefix("\n").rstrip(" \t").removesuffix("\n")
        for i, line in enumerate(re.split("\n", newline)):
            line = line.lstrip()
            if not line:
                nlines += 1
            if not line.startswith("#"):
                continue
            res.append(
                Token(
                    type=tokens.STANDALONE_COMMENT,
                    value=("\n" * nlines) + line,
                    line=t.line + i,  # type: ignore
                )
            )
            nlines = 0
        if res:
            res[-1].value += "\n" * nlines
        return res

    @staticmethod
    def _token_id(token: Token) -> Tuple[Any, ...]:
        return (
            token.value,
            token.type,
            token.column,
            token.end_pos,
            token.line,
        )

    def _generate_comment_associations(self, tree: Tree):
        """
        Run a BFS on the tree to find the final leaves of each lines
        where we can append the ignored comments in the final tree
        return a mapping of leaves to comment
        """

        # handle trailing comments first
        comments = self._comments
        comments.sort(key=lambda c: c.line)
        cmt_idx = {c.line: c for c in comments}

        # retrieve orphaned comments from ignored newlines
        # these are usually comments in between arguments to a call
        orphaned_comments = list(
            set(self._all_newlines) - set(self.indenter.processed_newlines)
        )
        orphaned_comments = [
            el
            for t in orphaned_comments
            for el in self._break_down_comments(t)
            if "#" in t
        ]
        orphaned_comments.sort(key=lambda c: c.line)
        o_cmt_idx = {c.line: c for c in orphaned_comments}

        queue = [tree]
        terminal_leaves: Dict[int, Token] = (
            {}
        )  # parent node and index of child

        # we first traverse to find the last token on each potentially relevant line
        while queue:
            node = queue.pop()
            # if the node does not cover any lines with comments
            # or is a leaf, we can skip
            if isinstance(node, Token):
                continue
            for i, child in enumerate(node.children):
                if (
                    isinstance(child, Token)
                    and (child.line and child.end_pos)
                    and (
                        child.line not in terminal_leaves
                        or child.end_pos  # type: ignore
                        >= terminal_leaves[child.line].end_pos
                    )
                ):
                    # we handle the case where a comment is the first node later
                    if child.type == NEWLINE and child.line == 1:
                        continue
                    else:
                        terminal_leaves[child.line] = child
                else:
                    queue.append(child)

        # create a mapping for trailing comments
        self._comment_mapping = {
            self._token_id(t): cmt_idx[line]
            for line, t in terminal_leaves.items()
            if line in cmt_idx
        }

        # we need to handle the case where the first line is a comment
        if comments and comments[0].line == 1 and 1 not in terminal_leaves:
            self._header_comments = [cmt_idx[1]]

        # handle the orphaned comments last
        for (
            line,
            comment,
        ) in o_cmt_idx.items():  # o_cmt_idx is sorted (p3.6+)
            attach_line = line
            while attach_line not in terminal_leaves:
                attach_line -= 1
                if attach_line < 0:
                    self._header_comments.append(comment)
                    break
            if attach_line < 0:
                continue
            t = terminal_leaves[attach_line]
            key_token = (t.value, t.type, t.column, t.end_pos, t.line)
            self._orphan_comment_mapping[key_token].append(comment)

    def _to_pytree(self, lark_tree: Tree) -> Node:
        def _transform(tree: Tree):
            """
            Convert a Lark tree to Pytree
            """
            subnodes = []
            for i, child in enumerate(tree.children):
                if isinstance(child, Tree):
                    subnodes.append(_transform(child))

                else:
                    # we store the original cursor of the subnodes
                    # for when we need to insert comments beforehand (whitespace)
                    subnode_cursor = len(subnodes)

                    if child.type in tokens.WHITESPACE and "#" in child.value:
                        subnodes.append(
                            Leaf(
                                type=child.type,
                                value="",
                            )
                        )
                        for comment in self._break_down_comments(child):
                            subnodes.append(
                                Leaf(
                                    type=tokens.STANDALONE_COMMENT,
                                    value=comment.value,
                                )
                            )

                    else:
                        # append current child to list of subnodes
                        subnodes.append(
                            Leaf(
                                type=child.type,
                                value=child.value,
                            )
                        )

                    # process any potential trailing comments
                    child_id = (
                        self._token_id(child)
                        if child.type
                        not in {tokens.COMMENT, tokens.STANDALONE_COMMENT}
                        else ("", "", 0, 0)
                    )

                    if (
                        self._comment_mapping
                        and child_id in self._comment_mapping
                    ):
                        comment = self._comment_mapping[child_id]
                        # if the current node is a newline, we want the trailing comment BEFORE
                        if child.type == NEWLINE:
                            subnodes.insert(
                                subnode_cursor,
                                Leaf(type=comment.type, value=comment.value),
                            )
                            subnode_cursor += 1
                        else:
                            subnodes.append(
                                Leaf(type=comment.type, value=comment.value)
                            )

                    # handle the standalone comments last
                    if (
                        self._orphan_comment_mapping
                        and child_id in self._orphan_comment_mapping
                    ):
                        subnodes += [
                            Leaf(type=c.type, value=c.value)
                            for c in self._orphan_comment_mapping[child_id]
                        ]

            node = Node(type=tree.data, children=[])
            for leaf in subnodes:
                node.append_child(leaf)
            return node

        module = _transform(lark_tree)

        self._header_comments.sort(key=lambda x: x.line, reverse=True)  # type: ignore
        for c in self._header_comments:
            module.children.insert(
                0,
                Leaf(
                    type=c.type,
                    value=c.value,
                ),
            )

        return module



# Punctuation
LSQB = "LSQB"
RSQB = "RSQB"
LPAR = "LPAR"
RPAR = "RPAR"
COMMA = "COMMA"
DOT = "DOT"
COLON = "COLON"
NAME = "NAME"
RBRACE = "RBRACE"
LBRACE = "LBRACE"
FLOAT_NUMBER = "FLOAT_NUMBER"
DEC_NUMBER = "DEC_NUMBER"
HEX_NUMBER = "HEX_NUMBER"
BIN_NUMBER = "BIN_NUMBER"
OCT_NUMBER = "OCT_NUMBER"
NUMBERS = {DEC_NUMBER, FLOAT_NUMBER, HEX_NUMBER, BIN_NUMBER, OCT_NUMBER}
AT = "AT"
BRACKET_MAP = {LPAR: RPAR, LSQB: RSQB, LBRACE: RBRACE}
OPENING_BRACKETS = set(BRACKET_MAP.keys())
CLOSING_BRACKETS = set(BRACKET_MAP.values())
BRACKETS = OPENING_BRACKETS | CLOSING_BRACKETS

# Formatting
INDENT = "_INDENT"
DEDENT = "_DEDENT"
NEWLINE = "_NEWLINE"
WHITESPACE = [INDENT, DEDENT, NEWLINE]

# Operations
AMPERSAND = "_BITAND"
VBAR = "_BITOR"
CARET = "_BITXOR"
BITAND = "bitand"
BITOR = "bitor"
BITXOR = "bitxor"
BITNOT = "invert"
PLUS = "PLUS"
MINUS = "MINUS"
PRODUCT = "_PRODUCT"
TILDE = "TILDE"
UNARY = {"usub", "uadd", "invert"}
ADD = "add"
SUB = "sub"
MUL = "mul"
MOD = "mod"
DIV = "div"
FLOORDIV = "floordiv"
POW = "pow"
SHL = "shl"
SHR = "shr"
RIGHTSHIFT = "_SHR"
LEFTSHIFT = "_SHL"
DOUBLESTAR = "_POW"
SLASH = "SLASH"
STAR = "WILDCARD"
PERCENT = "PERCENT"
OPERATIONS = {ADD, SUB, MUL, MOD, DIV, FLOORDIV}
EQUAL = "EQUAL"
ASSIGN = "assign"
AUG_ASSIGN = "aug_assign"
MULTIPLE_ASSIGN = "multiple_assign"
ASSIGN_OPERATORS = [
    ADD,
    SUB,
    DIV,
    FLOORDIV,
    MUL,
    MOD,
    POW,
    BITOR,
    BITXOR,
    BITAND,
    SHR,
    SHL,
]
MATH_OPERATORS = {
    PLUS,
    MINUS,
    STAR,
    SLASH,
    VBAR,
    AMPERSAND,
    CARET,
    PERCENT,
    LEFTSHIFT,
    RIGHTSHIFT,
    DOUBLESTAR,
}

BOOL_OR = "_OR"
BOOL_AND = "_AND"
BOOL_NOT = "_NOT"
LOGIC_OPERATORS = {"or", "and"}

COMPARATORS = {
    "LESSTHAN",
    "MORETHAN",
    "_EQ",
    "_NE",
    "_LE",
    "_GE",
}

# Imports
IMPORT_NAME = "_IMPORT"
IMPORT_FROM = "_FROM"
IMPORT = "import"

# Statements
PARAMETERS = "parameters"
ARGUMENTS = "arguments"
KWARG = "kwarg"
DECORATOR = "decorator"
FLOW_CONTROL = {"return", "break", "continue"}
INTERFACE_FUNCTION = "interface_function"
DECLARATIONS = {
    "_FUNC_DECL",
    "_EVENT_DECL",
    "_ENUM_DECL",
    "_STRUCT_DECL",
    "_INTERFACE_DECL",
}
DECLARATION = "declaration"
CONSTANT_DEF = "constant_def"
BODY = "body"
EVENT_BODY = "event_body"
ENUM_BODY = "enum_body"
MODULE = "module"
DEF_SUFFIX = "_def"
RETURN_TYPE = "_RETURN_TYPE"
FUNCTION_SIG = "function_sig"
ARRAY_DEF = "array_def"
DINARRAY_DEF = "dyn_array_def"
CALL = "call"
EXTERNAL_CALL = "external_call"
EMPTY = "empty"
ATTRIBUTE = "attribute"
GET_VAR = "get_var"
LOG_STMT = "log_stmt"
SUBSCRIPT = "subscript"
GETTER_SUFIX = "_with_getter"
CONSTANT = "constant"
IMMUTABLE = "immutable"
VARIABLE_ANNOTATION = "variable_annotation"
ABI_DECODE = "abi_decode"
PUBLIC_MODIFIER = "public_modifier"
REENTRANT_MODIFIER = "reentrant_modifier"
IMMUTABLE_MODIFIER = "immutable_modifier"
TRANSIENT_MODIFIER = "transient_modifier"
IMPLEMENTS = "implements"
USES = "uses"
INDEXED_ARGS = "indexed_event_arg"
ATOM = "atom"

COND_EXEC = "cond_exec"
FOR_STMT = "for_stmt"
IF_STMT = "if_stmt"
TERNARY = "ternary"
FOR = "FOR"
IF = "IF"
IN = "_IN"
ELSE = "ELSE"
ELIF = "ELIF"

PRAGMA = "PRAGMA"
COMMENT = "COMMENT"
STRING = "STRING"
DOCSTRING = "DOCSTRING"
STANDALONE_COMMENT = "STANDALONE_COMMENT"

ASSERT = "assert"
ASSERT_TOKEN = "_ASSERT"
ASSERT_WITH_REASON = "assert_with_reason"
ASSERT_UNREACHABLE = "assert_unreachable"
ASSERTS = {ASSERT, ASSERT_WITH_REASON, ASSERT_UNREACHABLE}

VARIABLE_DEF = "variable_def"
RETURN_STMT = "return_stmt"
PASS_STMT = "pass_stmt"
BREAK_STMT = "break_stmt"
CONTINUE_STMT = "continue_stmt"
INITIALIZES_STMT = "initializes_stmt"
IMPORTED_TYPE = "imported_type"
RAISE = "raise"
RAISE_WITH_REASON = "raise_with_reason"
IMMUTABLE_DEF = "immutable_def"
INTERFACE_DEF = "interface_def"
IMPLEMENTS_DEF = "implements_def"
USES_DEF = "uses_def"
EXPORT = "export"
STRUCT_DEF = "struct_def"
ENUM_DEF = "enum_def"
EVENT_DEF = "event_def"
INDEXED_EVENT_ARG = "indexed_event_arg"
EVENT_MEMBER = "event_member"
ENUM_MEMBER = "enum_member"
STRUCT_MEMBER = "struct_member"

BODIES = {BODY, EVENT_BODY, ENUM_BODY, MODULE, STRUCT_DEF}

SIMPLE_STATEMENTS = {
    VARIABLE_DEF,
    RETURN_STMT,
    PASS_STMT,
    BREAK_STMT,
    CONTINUE_STMT,
    RAISE,
    RAISE_WITH_REASON,
    LOG_STMT,
    INITIALIZES_STMT,
    IMPLEMENTS_DEF,
    USES_DEF,
    EXPORT,
    CONSTANT_DEF,
    IMMUTABLE_DEF,
    INTERFACE_DEF,
    STRUCT_DEF,
    ENUM_DEF,
    EVENT_DEF,
    INDEXED_EVENT_ARG,
    EVENT_MEMBER,
    ENUM_MEMBER,
    STRUCT_MEMBER,
    ASSIGN,
    AUG_ASSIGN,
} | ASSERTS

ASSIGNMENTS_SIGNS = {"=", "+=", "-=", "*=", "/=", "%=", "&=", "|=", "^="}
